"""
PDL to PDL_interp Transformation
"""

from abc import ABC
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import Optional, cast

from xdsl.builder import Builder
from xdsl.dialects import pdl, pdl_interp
from xdsl.dialects.builtin import (
    ArrayAttr,
    IntegerAttr,
    ModuleOp,
    StringAttr,
    TypeAttribute,
)
from xdsl.ir import (
    Block,
    Operation,
    OpResult,
    Region,
    SSAValue,
)
from xdsl.rewriter import InsertPoint
from xdsl.transforms.convert_pdl_to_pdl_interp.predicate import (
    Answer,
    AttributeAnswer,
    AttributeConstraintQuestion,
    AttributeLiteralPosition,
    AttributePosition,
    ConstraintPosition,
    ConstraintQuestion,
    EqualToQuestion,
    ForEachPosition,
    IsNotNullQuestion,
    OperandCountAtLeastQuestion,
    OperandCountQuestion,
    OperandGroupPosition,
    OperandPosition,
    OperationNameQuestion,
    OperationPosition,
    Position,
    PositionalPredicate,
    Predicate,
    Question,
    ResultCountAtLeastQuestion,
    ResultCountQuestion,
    ResultGroupPosition,
    ResultPosition,
    StringAnswer,
    TrueAnswer,
    TypeAnswer,
    TypeConstraintQuestion,
    TypeLiteralPosition,
    TypePosition,
    UnsignedAnswer,
    UsersPosition,
    get_position_cost,
    get_question_cost,
)
from xdsl.utils.scoped_dict import ScopedDict

# =============================================================================
# Matcher Tree Nodes
# =============================================================================


@dataclass
class MatcherNode(ABC):
    """Base class for matcher tree nodes"""

    position: Position | None = None
    question: Question | None = None
    failure_node: Optional["MatcherNode"] = None


@dataclass(kw_only=True)
class BoolNode(MatcherNode):
    """Boolean predicate node"""

    success_node: MatcherNode | None = None
    failure_node: MatcherNode | None = None

    answer: Answer


@dataclass
class SwitchNode(MatcherNode):
    """Multi-way switch node"""

    children: dict[Answer, MatcherNode | None] = field(default_factory=lambda: {})


@dataclass(kw_only=True)
class SuccessNode(MatcherNode):
    """Successful pattern match"""

    pattern: pdl.PatternOp  # PDL pattern reference
    root: SSAValue | None = None  # Root value


@dataclass
class ExitNode(MatcherNode):
    """Exit/failure node"""

    pass


# =============================================================================
# Pattern Analysis
# =============================================================================


class PatternAnalyzer:
    """Analyzes PDL patterns and extracts predicates"""

    def detect_roots(self, pattern: pdl.PatternOp) -> list[OpResult[pdl.OperationType]]:
        """Detect root operations in a pattern"""
        used = {
            operand.owner.parent_
            for operation_op in pattern.body.ops
            if isinstance(operation_op, pdl.OperationOp)
            for operand in operation_op.operand_values
            if isinstance(operand.owner, pdl.ResultOp | pdl.ResultsOp)
        }

        rewriter = pattern.body.block.last_op
        assert isinstance(rewriter, pdl.RewriteOp)
        if rewriter.root is not None:
            if rewriter.root in used:
                used.remove(rewriter.root)

        roots = [
            op.op
            for op in pattern.body.ops
            if isinstance(op, pdl.OperationOp) and op.op not in used
        ]
        return roots

    def extract_tree_predicates(
        self,
        value: SSAValue,
        position: Position,
        inputs: dict[SSAValue, Position],
        ignore_operand: int | None = None,
    ) -> list[PositionalPredicate]:
        """Extract predicates by walking the operation tree"""
        predicates: list[PositionalPredicate] = []

        # Check if this value has been visited before
        existing_pos = inputs.get(value)
        if existing_pos is not None:
            # If this is an input value that has been visited in the tree,
            # add a constraint to ensure both instances refer to the same value
            defining_op = value.owner
            if isinstance(
                defining_op,
                pdl.AttributeOp
                | pdl.OperandOp
                | pdl.OperandsOp
                | pdl.OperationOp
                | pdl.TypeOp
                | pdl.TypesOp,
            ):
                # Order positions by depth (deeper position gets the equality predicate)
                if position.get_operation_depth() > existing_pos.get_operation_depth():
                    deeper_pos, shallower_pos = position, existing_pos
                else:
                    deeper_pos, shallower_pos = existing_pos, position

                equal_pred = Predicate.get_equal_to(shallower_pos)
                predicates.append(
                    PositionalPredicate(
                        q=equal_pred.q, a=equal_pred.a, position=deeper_pos
                    )
                )
            return predicates

        inputs[value] = position

        # Dispatch based on position type (not value type!)
        match position:
            case AttributePosition():
                assert isinstance(value, OpResult)
                predicates.extend(
                    self._extract_attribute_predicates(value.owner, position, inputs)
                )
            case OperationPosition():
                assert isinstance(value, OpResult)
                predicates.extend(
                    self._extract_operation_predicates(
                        value.owner, position, inputs, ignore_operand
                    )
                )
            case TypePosition():
                assert isinstance(value, OpResult)
                predicates.extend(
                    self._extract_type_predicates(value.owner, position, inputs)
                )
            case OperandPosition() | OperandGroupPosition():
                assert isinstance(value, SSAValue)
                predicates.extend(
                    self._extract_operand_tree_predicates(value, position, inputs)
                )
            case _:
                raise TypeError(f"Unexpected position kind: {type(position)}")

        return predicates

    def _get_num_non_range_values(self, values: Sequence[SSAValue]) -> int:
        """Returns the number of non-range elements within values"""
        return sum(1 for v in values if not isinstance(v.type, pdl.RangeType))

    def _extract_attribute_predicates(
        self,
        attr_op: Operation,
        attr_pos: AttributePosition,
        inputs: dict[SSAValue, Position],
    ) -> list[PositionalPredicate]:
        """Extract predicates for an attribute"""
        predicates: list[PositionalPredicate] = []

        is_not_null = Predicate.get_is_not_null()
        predicates.append(
            PositionalPredicate(q=is_not_null.q, a=is_not_null.a, position=attr_pos)
        )

        if isinstance(attr_op, pdl.AttributeOp):
            if attr_op.value_type:
                type_pos = attr_pos.get_type()
                predicates.extend(
                    self.extract_tree_predicates(attr_op.value_type, type_pos, inputs)
                )

            elif attr_op.value:
                attr_constraint = Predicate.get_attribute_constraint(attr_op.value)
                predicates.append(
                    PositionalPredicate(
                        q=attr_constraint.q, a=attr_constraint.a, position=attr_pos
                    )
                )

        return predicates

    def _extract_operation_predicates(
        self,
        op_op: Operation,
        op_pos: OperationPosition,
        inputs: dict[SSAValue, Position],
        ignore_operand: int | None = None,
    ) -> list[PositionalPredicate]:
        """Extract predicates for an operation"""
        predicates: list[PositionalPredicate] = []

        if not isinstance(op_op, pdl.OperationOp):
            return predicates

        if not op_pos.is_root():
            is_not_null = Predicate.get_is_not_null()
            predicates.append(
                PositionalPredicate(q=is_not_null.q, a=is_not_null.a, position=op_pos)
            )

        # Operation name check
        if op_op.opName:
            op_name = op_op.opName.data
            op_name_pred = Predicate.get_operation_name(op_name)
            predicates.append(
                PositionalPredicate(q=op_name_pred.q, a=op_name_pred.a, position=op_pos)
            )

        operands = op_op.operand_values
        min_operands = self._get_num_non_range_values(operands)
        if min_operands != len(operands):
            # Has variadic operands - check minimum
            if min_operands > 0:
                operand_count_pred = Predicate.get_operand_count_at_least(min_operands)
                predicates.append(
                    PositionalPredicate(
                        q=operand_count_pred.q, a=operand_count_pred.a, position=op_pos
                    )
                )
        else:
            # All non-variadic - check exact count
            operand_count_pred = Predicate.get_operand_count(min_operands)
            predicates.append(
                PositionalPredicate(
                    q=operand_count_pred.q, a=operand_count_pred.a, position=op_pos
                )
            )

        types = op_op.type_values
        min_results = self._get_num_non_range_values(types)
        if min_results == len(types):
            # All non-variadic - check exact count
            result_count_pred = Predicate.get_result_count(len(types))
            predicates.append(
                PositionalPredicate(
                    q=result_count_pred.q, a=result_count_pred.a, position=op_pos
                )
            )
        elif min_results > 0:
            # Has variadic results - check minimum
            result_count_pred = Predicate.get_result_count_at_least(min_results)
            predicates.append(
                PositionalPredicate(
                    q=result_count_pred.q, a=result_count_pred.a, position=op_pos
                )
            )

        # Process attributes
        for attr_name, attr in zip(
            op_op.attributeValueNames, op_op.attribute_values, strict=True
        ):
            attr_pos = op_pos.get_attribute(attr_name.data)
            predicates.extend(self.extract_tree_predicates(attr, attr_pos, inputs))

        if len(operands) == 1 and isinstance(operands[0].type, pdl.RangeType):
            # Special case: single variadic operand represents all operands
            if op_pos.is_root() or op_pos.is_operand_defining_op():
                all_operands_pos = op_pos.get_all_operands()
                predicates.extend(
                    self.extract_tree_predicates(operands[0], all_operands_pos, inputs)
                )
        else:
            # Process individual operands
            found_variable_length = False
            for i, operand in enumerate(operands):
                is_variadic = isinstance(operand.type, pdl.RangeType)
                found_variable_length = found_variable_length or is_variadic

                if ignore_operand is not None and i == ignore_operand:
                    continue

                # Switch to group-based positioning after first variadic
                if found_variable_length:
                    operand_pos = op_pos.get_operand_group(i, is_variadic)
                else:
                    operand_pos = op_pos.get_operand(i)

                predicates.extend(
                    self.extract_tree_predicates(operand, operand_pos, inputs)
                )

        if len(types) == 1 and isinstance(types[0].type, pdl.RangeType):
            # Single variadic result represents all results
            all_results_pos = op_pos.get_all_results()
            type_pos = all_results_pos.get_type()
            predicates.extend(self.extract_tree_predicates(types[0], type_pos, inputs))
        else:
            # Process individual results
            found_variable_length = False
            for i, type_value in enumerate(types):
                is_variadic = isinstance(type_value.type, pdl.RangeType)
                found_variable_length = found_variable_length or is_variadic

                # Switch to group-based positioning after first variadic
                if found_variable_length:
                    result_pos = op_pos.get_result_group(i, is_variadic)
                else:
                    result_pos = op_pos.get_result(i)

                # Add not-null check for each result
                is_not_null = Predicate.get_is_not_null()
                predicates.append(
                    PositionalPredicate(
                        q=is_not_null.q, a=is_not_null.a, position=result_pos
                    )
                )

                # Process the result type
                type_pos = result_pos.get_type()
                predicates.extend(
                    self.extract_tree_predicates(type_value, type_pos, inputs)
                )

        return predicates

    def _extract_operand_tree_predicates(
        self,
        operand_value: SSAValue,
        operand_pos: OperandPosition | OperandGroupPosition,
        inputs: dict[SSAValue, Position],
    ) -> list[PositionalPredicate]:
        """Extract predicates for an operand or operand group"""
        predicates: list[PositionalPredicate] = []

        # Get the defining operation
        defining_op = operand_value.owner
        is_variadic = isinstance(operand_value.type, pdl.RangeType)

        match defining_op:
            case pdl.OperandOp() | pdl.OperandsOp():
                match defining_op:
                    case pdl.OperandOp():
                        is_not_null = Predicate.get_is_not_null()
                        predicates.append(
                            PositionalPredicate(
                                q=is_not_null.q, a=is_not_null.a, position=operand_pos
                            )
                        )
                    case pdl.OperandsOp() if (
                        isinstance(operand_pos, OperandGroupPosition)
                        and operand_pos.group_number is not None
                    ):
                        is_not_null = Predicate.get_is_not_null()
                        predicates.append(
                            PositionalPredicate(
                                q=is_not_null.q, a=is_not_null.a, position=operand_pos
                            )
                        )
                    case _:
                        pass

                if defining_op.value_type:
                    type_pos = operand_pos.get_type()
                    predicates.extend(
                        self.extract_tree_predicates(
                            defining_op.value_type, type_pos, inputs
                        )
                    )

            case pdl.ResultOp() | pdl.ResultsOp():
                index_attr = defining_op.index
                index = index_attr.value.data if index_attr is not None else None

                if index is not None:
                    is_not_null = Predicate.get_is_not_null()
                    predicates.append(
                        PositionalPredicate(
                            q=is_not_null.q, a=is_not_null.a, position=operand_pos
                        )
                    )

                # Get the parent operation position
                parent_op = defining_op.parent_
                defining_op_pos = operand_pos.get_defining_op()

                # Parent operation should not be null
                is_not_null = Predicate.get_is_not_null()
                predicates.append(
                    PositionalPredicate(
                        q=is_not_null.q, a=is_not_null.a, position=defining_op_pos
                    )
                )

                match defining_op:
                    case pdl.ResultOp():
                        result_pos = defining_op_pos.get_result(
                            index if index is not None else 0
                        )
                    case pdl.ResultsOp():  # ResultsOp
                        result_pos = defining_op_pos.get_result_group(
                            index, is_variadic
                        )

                equal_to = Predicate.get_equal_to(operand_pos)
                predicates.append(
                    PositionalPredicate(q=equal_to.q, a=equal_to.a, position=result_pos)
                )

                # Recursively process the parent operation
                predicates.extend(
                    self.extract_tree_predicates(parent_op, defining_op_pos, inputs)
                )
            case _:
                pass

        return predicates

    def _extract_type_predicates(
        self,
        type_op: Operation,
        type_pos: TypePosition,
        inputs: dict[SSAValue, Position],
    ) -> list[PositionalPredicate]:
        """Extract predicates for a type"""
        predicates: list[PositionalPredicate] = []

        match type_op:
            case pdl.TypeOp(constantType=const_type) if const_type:
                type_constraint = Predicate.get_type_constraint(const_type)
                predicates.append(
                    PositionalPredicate(
                        q=type_constraint.q, a=type_constraint.a, position=type_pos
                    )
                )
            case pdl.TypesOp(constantTypes=const_types) if const_types:
                type_constraint = Predicate.get_type_constraint(const_types)
                predicates.append(
                    PositionalPredicate(
                        q=type_constraint.q, a=type_constraint.a, position=type_pos
                    )
                )
            case _:
                pass

        return predicates

    def extract_non_tree_predicates(
        self,
        pattern: pdl.PatternOp,
        inputs: dict[SSAValue, Position],
    ) -> list[PositionalPredicate]:
        """Extract predicates that cannot be determined via tree walking"""
        predicates: list[PositionalPredicate] = []

        for op in pattern.body.ops:
            match op:
                case pdl.AttributeOp():
                    if op.output not in inputs:
                        if op.value:
                            # Create literal position for constant attribute
                            attr_pos = AttributeLiteralPosition(
                                value=op.value, parent=None
                            )
                            inputs[op.output] = attr_pos

                case pdl.ApplyNativeConstraintOp():
                    # Collect all argument positions
                    arg_positions = tuple(inputs.get(arg) for arg in op.args)
                    for pos in arg_positions:
                        assert pos is not None
                    arg_positions = cast(tuple[Position, ...], arg_positions)

                    # Find the furthest position (deepest)
                    furthest_pos = max(
                        arg_positions, key=lambda p: p.get_operation_depth() if p else 0
                    )

                    # Create the constraint predicate
                    result_types = tuple(r.type for r in op.res)
                    is_negated = bool(op.is_negated.value.data)
                    constraint_pred = Predicate.get_constraint(
                        op.constraint_name.data, arg_positions, result_types, is_negated
                    )

                    # Register positions for constraint results
                    for i, result in enumerate(op.results):
                        assert isinstance(constraint_pred.q, ConstraintQuestion)
                        constraint_pos = ConstraintPosition.get_constraint(
                            constraint_pred.q, i
                        )
                        existing = inputs.get(result)
                        if existing:
                            # Add equality constraint if result already has a position
                            deeper, shallower = (
                                (constraint_pos, existing)
                                if furthest_pos.get_operation_depth()
                                > existing.get_operation_depth()
                                else (existing, constraint_pos)
                            )
                            eq_pred = Predicate.get_equal_to(shallower)
                            predicates.append(
                                PositionalPredicate(
                                    q=eq_pred.q, a=eq_pred.a, position=deeper
                                )
                            )
                        else:
                            inputs[result] = constraint_pos

                    predicates.append(
                        PositionalPredicate(
                            q=constraint_pred.q,
                            a=constraint_pred.a,
                            position=furthest_pos,
                        )
                    )

                case pdl.ResultOp():
                    # Ensure result exists
                    if op.val not in inputs:
                        assert isinstance(op.parent_.owner, pdl.OperationOp)
                        parent_pos = inputs.get(op.parent_.owner.op)
                        if parent_pos and isinstance(parent_pos, OperationPosition):
                            result_pos = parent_pos.get_result(op.index.value.data)
                            is_not_null = Predicate.get_is_not_null()
                            predicates.append(
                                PositionalPredicate(
                                    q=is_not_null.q,
                                    a=is_not_null.a,
                                    position=result_pos,
                                )
                            )

                case pdl.ResultsOp():
                    # Handle result groups
                    if op.val not in inputs:
                        assert isinstance(op.parent_.owner, pdl.OperationOp)
                        parent_pos = inputs.get(op.parent_.owner.op)
                        if parent_pos and isinstance(parent_pos, OperationPosition):
                            is_variadic = isinstance(op.val.type, pdl.RangeType)
                            index = op.index.value.data if op.index else None
                            result_pos = parent_pos.get_result_group(index, is_variadic)
                            if index is not None:
                                is_not_null = Predicate.get_is_not_null()
                                predicates.append(
                                    PositionalPredicate(
                                        q=is_not_null.q,
                                        a=is_not_null.a,
                                        position=result_pos,
                                    )
                                )

                case pdl.TypeOp():
                    # Handle constant types
                    if op.result not in inputs and op.constantType:
                        type_pos = TypeLiteralPosition.get_type_literal(
                            value=op.constantType
                        )
                        inputs[op.result] = type_pos

                case pdl.TypesOp():
                    # Handle constant type arrays
                    if op.result not in inputs and op.constantTypes:
                        type_pos = TypeLiteralPosition.get_type_literal(
                            value=op.constantTypes
                        )
                        inputs[op.result] = type_pos

                case _:
                    pass

        return predicates


# =============================================================================
# Predicate Ordering and Tree Construction
# =============================================================================


@dataclass
class OrderedPredicate:
    """Predicate with ordering information"""

    position: Position
    question: Question
    primary_score: int = 0  # Frequency across patterns
    secondary_score: int = 0  # Squared sum within patterns
    tie_breaker: int = 0  # Insertion order
    pattern_answers: dict[pdl.PatternOp, Answer] = field(default_factory=lambda: {})

    def __lt__(self, other: "OrderedPredicate") -> bool:
        """Comparison for priority ordering"""
        return (
            self.primary_score,
            self.secondary_score,
            -self.position.get_operation_depth(),  # Prefer lower depth
            -get_position_cost(self.position),  # Position dependency
            -get_question_cost(self.question),  # Predicate dependency
            -self.tie_breaker,  # Deterministic order
        ) > (
            other.primary_score,
            other.secondary_score,
            -other.position.get_operation_depth(),
            -get_position_cost(other.position),
            -get_question_cost(other.question),
            -other.tie_breaker,
        )

    def __hash__(self):
        """The hash is based on the immutable identity of the predicate."""
        return hash((self.position, self.question))


def _depends_on(pred_a: OrderedPredicate, pred_b: OrderedPredicate) -> bool:
    """Returns true if predicate 'b' depends on a result of predicate 'a'."""
    constraint_q_a = pred_a.question
    if not isinstance(constraint_q_a, ConstraintQuestion):
        return False

    def position_depends_on_a(pos: Position) -> bool:
        if isinstance(pos, ConstraintPosition):
            return pos.constraint == constraint_q_a
        return False

    if isinstance(pred_b.question, ConstraintQuestion):
        # Does any argument of b use a?
        return any(position_depends_on_a(arg) for arg in pred_b.question.arg_positions)
    if isinstance(pred_b.question, EqualToQuestion):
        return position_depends_on_a(pred_b.position) or position_depends_on_a(
            pred_b.question.other_position
        )
    return position_depends_on_a(pred_b.position)


def _stable_topological_sort(
    predicates: list[OrderedPredicate],
) -> list[OrderedPredicate]:
    """Sorts predicates topologically while maintaining stability for independent items."""
    # Build dependency graph
    dependencies: dict[OrderedPredicate, set[OrderedPredicate]] = {
        p: set() for p in predicates
    }
    for i, pred_b in enumerate(predicates):
        for j in range(i + 1, len(predicates)):
            pred_a = predicates[j]
            if _depends_on(pred_a, pred_b):
                dependencies[pred_b].add(pred_a)  # b depends on a

    sorted_list: list[OrderedPredicate] = []
    pred_list: list[OrderedPredicate] = predicates[:]

    while pred_list:
        # Find all items with no dependencies within the current list
        to_sort = [
            p for p in pred_list if all(dep not in pred_list for dep in dependencies[p])
        ]
        # It is not possible to have cycles in the dependency graph
        # because predicates can only depend on predicates containing
        # a ConstraintQuestion. It is however impossible to construct
        # a cycle with those because they are frozen at construction
        # and thus cannot refer to a ConstraintPosition that was
        # defined later.
        assert to_sort, "Encountered a cycle!"

        # Append them to the sorted list
        sorted_list.extend(to_sort)

        # Remove them from the list to be sorted
        pred_list = [p for p in pred_list if p not in to_sort]

    return sorted_list


class PredicateTreeBuilder:
    """Builds optimized predicate matching trees"""

    analyzer: PatternAnalyzer
    _pattern_roots: dict[pdl.PatternOp, SSAValue]
    pattern_value_positions: dict[pdl.PatternOp, dict[SSAValue, Position]]

    def __init__(self):
        self.analyzer = PatternAnalyzer()
        self._pattern_roots = {}
        self.pattern_value_positions = {}

    def build_predicate_tree(self, patterns: list[pdl.PatternOp]) -> MatcherNode:
        """Build optimized matcher tree from multiple patterns"""

        # Extract predicates for all patterns
        all_pattern_predicates: list[
            tuple[pdl.PatternOp, list[PositionalPredicate]]
        ] = []
        for pattern in patterns:
            predicates, root, inputs = self._extract_pattern_predicates(pattern)
            all_pattern_predicates.append((pattern, predicates))
            self._pattern_roots[pattern] = root
            self.pattern_value_positions[pattern] = inputs

        # Create ordered predicates with frequency analysis
        ordered_predicates = self._create_ordered_predicates(all_pattern_predicates)
        # Sort predicates by priority
        sorted_predicates = sorted(ordered_predicates.values())
        sorted_predicates = _stable_topological_sort(sorted_predicates)

        # Build matcher tree by propagating patterns
        root_node = None
        for pattern, predicates in all_pattern_predicates:
            if not predicates:
                continue
            pattern_predicate_set = {
                (pred.position, pred.q): pred for pred in predicates
            }
            root_node = self._propagate_pattern(
                root_node, pattern, pattern_predicate_set, sorted_predicates, 0
            )

        # Add exit node and optimize
        if root_node is not None:
            root_node = self._optimize_tree(root_node)
            root_node = self._insert_exit_node(root_node)
            return root_node
        else:
            # Return a default exit node if no patterns were processed
            return ExitNode()

    def _extract_pattern_predicates(
        self, pattern: pdl.PatternOp
    ) -> tuple[list[PositionalPredicate], SSAValue, dict[SSAValue, Position]]:
        """Extract all predicates for a single pattern"""
        predicates: list[PositionalPredicate] = []
        inputs: dict[SSAValue, Position] = {}

        roots = self.analyzer.detect_roots(pattern)
        if len(roots) != 1:
            raise ValueError("Multi-root patterns are not yet supported.")

        rewriter = pattern.body.block.last_op
        assert isinstance(rewriter, pdl.RewriteOp)
        best_root = rewriter.root if rewriter.root is not None else roots[0]

        # Downward traversal from the best root
        root_pos = OperationPosition(depth=0)
        predicates.extend(
            self.analyzer.extract_tree_predicates(best_root, root_pos, inputs)
        )

        predicates.extend(self.analyzer.extract_non_tree_predicates(pattern, inputs))
        return predicates, best_root, inputs

    def _create_ordered_predicates(
        self,
        all_pattern_predicates: list[tuple[pdl.PatternOp, list[PositionalPredicate]]],
    ) -> dict[tuple[Position, Question], OrderedPredicate]:
        """Create ordered predicates with frequency analysis"""
        predicate_map: dict[tuple[Position, Question], OrderedPredicate] = {}
        tie_breaker = 0

        # Collect unique predicates
        for pattern, predicates in all_pattern_predicates:
            for pred in predicates:
                key = (pred.position, pred.q)

                if key not in predicate_map:
                    ordered_pred = OrderedPredicate(
                        position=pred.position,
                        question=pred.q,
                        tie_breaker=tie_breaker,
                    )
                    predicate_map[key] = ordered_pred
                    tie_breaker += 1

                # Track pattern answers and increment frequency
                predicate_map[key].pattern_answers[pattern] = pred.a
                predicate_map[key].primary_score += 1

        # Calculate secondary scores
        for pattern, predicates in all_pattern_predicates:
            pattern_primary_sum = 0
            seen_keys: set[tuple[Position, Question]] = (
                set()
            )  # Track unique keys per pattern

            # First pass: collect unique predicates for this pattern
            for pred in predicates:
                key = (pred.position, pred.q)
                if key not in seen_keys:
                    seen_keys.add(key)
                    ordered_pred = predicate_map[key]
                    pattern_primary_sum += ordered_pred.primary_score**2

            # Second pass: add secondary score to each unique predicate
            for key in seen_keys:
                ordered_pred = predicate_map[key]
                ordered_pred.secondary_score += pattern_primary_sum

        return predicate_map

    def _propagate_pattern(
        self,
        node: MatcherNode | None,
        pattern: pdl.PatternOp,
        pattern_predicates: dict[tuple[Position, Question], PositionalPredicate],
        sorted_predicates: list[OrderedPredicate],
        predicate_index: int,
    ) -> MatcherNode:
        """Propagate a pattern through the predicate tree"""

        # Base case: reached end of predicates
        if predicate_index >= len(sorted_predicates):
            root_val = self._pattern_roots.get(pattern)
            return SuccessNode(pattern=pattern, root=root_val, failure_node=node)

        current_predicate = sorted_predicates[predicate_index]
        pred_key = (current_predicate.position, current_predicate.question)

        # Skip predicates not in this pattern
        if pred_key not in pattern_predicates:
            return self._propagate_pattern(
                node,
                pattern,
                pattern_predicates,
                sorted_predicates,
                predicate_index + 1,
            )

        # Create or match existing node
        if node is None:
            # Create new switch node
            node = SwitchNode(
                position=current_predicate.position, question=current_predicate.question
            )

        if self._nodes_match(node, current_predicate):
            # Continue down matching path
            pattern_answer = pattern_predicates[pred_key].a

            if isinstance(node, SwitchNode):
                if pattern_answer not in node.children:
                    node.children[pattern_answer] = None

                node.children[pattern_answer] = self._propagate_pattern(
                    node.children[pattern_answer],
                    pattern,
                    pattern_predicates,
                    sorted_predicates,
                    predicate_index + 1,
                )

        else:
            # Divergence - continue down failure path
            node.failure_node = self._propagate_pattern(
                node.failure_node,
                pattern,
                pattern_predicates,
                sorted_predicates,
                predicate_index,
            )

        return node

    def _nodes_match(self, node: MatcherNode, predicate: OrderedPredicate) -> bool:
        """Check if node matches the given predicate"""
        return (
            node.position == predicate.position and node.question == predicate.question
        )

    def _insert_exit_node(self, root: MatcherNode) -> MatcherNode:
        """Insert exit node at end of failure paths"""
        curr = root
        while curr.failure_node:
            curr = curr.failure_node
        curr.failure_node = ExitNode()
        return root

    def _optimize_tree(self, root: MatcherNode) -> MatcherNode:
        """Optimize the tree by collapsing single-child switches to bools"""
        # Recursively optimize children
        if isinstance(root, SwitchNode):
            for answer in root.children:
                child_node = root.children[answer]
                if child_node is not None:
                    root.children[answer] = self._optimize_tree(child_node)
        elif isinstance(root, BoolNode):
            if root.success_node is not None:
                root.success_node = self._optimize_tree(root.success_node)

        if root.failure_node is not None:
            root.failure_node = self._optimize_tree(root.failure_node)

        if isinstance(root, SwitchNode) and len(root.children) == 1:
            # Convert switch to bool node
            answer, child = next(iter(root.children.items()))
            bool_node = BoolNode(
                position=root.position,
                question=root.question,
                success_node=child,
                failure_node=root.failure_node,
                answer=answer,
            )
            return bool_node

        return root


# =============================================================================
# Code Generation
# =============================================================================


class MatcherGenerator:
    """Generates PDL interpreter matcher from matcher tree"""

    matcher_func: pdl_interp.FuncOp
    rewriter_module: ModuleOp
    rewriter_builder: Builder
    value_to_position: dict[pdl.PatternOp, dict[SSAValue, Position]]
    values: ScopedDict[Position, SSAValue]
    failure_block_stack: list[Block]
    builder: Builder
    constraint_op_map: dict[ConstraintQuestion, pdl_interp.ApplyConstraintOp]
    rewriter_names: dict[str, int]

    def __init__(
        self,
        matcher_func: pdl_interp.FuncOp,
        rewriter_module: ModuleOp,
        optimize_for_eqsat: bool = False,
    ) -> None:
        self.matcher_func = matcher_func
        self.rewriter_module = rewriter_module
        self.rewriter_builder = Builder(InsertPoint.at_end(rewriter_module.body.block))
        self.value_to_position = {}
        self.values = ScopedDict()
        self.failure_block_stack = []
        self.builder = Builder(InsertPoint.at_start(matcher_func.body.block))
        self.constraint_op_map = {}
        self.rewriter_names = {}

    def generate_matcher(
        self, node: MatcherNode, region: Region, block: Block | None = None
    ) -> Block:
        """Generate PDL interpreter operations for a matcher node"""

        # Create block if needed
        if block is None:
            block = Block()
            region.add_block(block)

        # Handle exit node - just add finalize
        if isinstance(node, ExitNode):
            finalize_op = pdl_interp.FinalizeOp()
            self.builder.insert_op(finalize_op, InsertPoint.at_end(block))
            return block

        self.values = ScopedDict(self.values)
        assert self.values.parent is not None

        # Handle failure node
        failure_block = None
        if node.failure_node:
            failure_block = self.generate_matcher(node.failure_node, region)
            self.failure_block_stack.append(failure_block)
        else:
            assert self.failure_block_stack, "Expected valid failure block"
            failure_block = self.failure_block_stack[-1]

        # Get value for position if exists
        current_block = block
        val = None
        if node.position:
            val = self.get_value_at(current_block, node.position)

        # Dispatch based on node type
        match node:
            case BoolNode():
                assert val is not None
                self.generate_bool_node(node, current_block, val)
            case SwitchNode():
                assert val is not None
                self.generate_switch_node(node, current_block, val)
            case SuccessNode():
                raise NotImplementedError()
            case _:
                raise NotImplementedError(f"Unhandled node type {type(node)}")

        # Pop failure block if we pushed one
        if node.failure_node:
            self.failure_block_stack.pop()

        self.values = self.values.parent  # Pop scope
        return block

    def get_value_at(self, block: Block, position: Position) -> SSAValue:
        """Get or create SSA value for a position"""

        # Check cache
        if position in self.values:
            return self.values[position]

        # Get parent value if needed
        parent_val = None
        if position.parent:
            parent_val = self.get_value_at(block, position.parent)

        # Create value based on position type
        self.builder.insertion_point = InsertPoint.at_end(block)
        value = None

        if isinstance(position, OperationPosition):
            if position.is_operand_defining_op():
                assert parent_val is not None
                # Get defining operation of operand
                defining_op = pdl_interp.GetDefiningOpOp(parent_val)
                defining_op.attributes["position"] = StringAttr(position.__repr__())
                self.builder.insert(defining_op)
                value = defining_op.input_op
            else:
                # Passthrough
                value = parent_val

        elif isinstance(position, OperandPosition):
            assert parent_val is not None
            get_operand_op = pdl_interp.GetOperandOp(
                position.operand_number, parent_val
            )
            self.builder.insert(get_operand_op)
            value = get_operand_op.value

        elif isinstance(position, OperandGroupPosition):
            assert parent_val is not None
            # Get operands (possibly variadic)
            result_type = (
                pdl.RangeType(pdl.ValueType())
                if position.is_variadic
                else pdl.ValueType()
            )
            raise NotImplementedError("pdl_interp.get_operands is not yet implemented")
            get_operands_op = pdl_interp.GetOperandsOp(
                position.group_number, parent_val, result_type
            )
            self.builder.insert(get_operands_op)
            value = get_operands_op.value

        elif isinstance(position, ResultPosition):
            assert parent_val is not None
            get_result_op = pdl_interp.GetResultOp(position.result_number, parent_val)
            self.builder.insert(get_result_op)
            value = get_result_op.value

        elif isinstance(position, ResultGroupPosition):
            assert parent_val is not None
            # Get results (possibly variadic)
            result_type = (
                pdl.RangeType(pdl.ValueType())
                if position.is_variadic
                else pdl.ValueType()
            )
            get_results_op = pdl_interp.GetResultsOp(
                position.group_number, parent_val, result_type
            )
            self.builder.insert(get_results_op)
            value = get_results_op.value

        elif isinstance(position, AttributePosition):
            assert parent_val is not None
            get_attr_op = pdl_interp.GetAttributeOp(position.attribute_name, parent_val)
            self.builder.insert(get_attr_op)
            value = get_attr_op.value

        elif isinstance(position, AttributeLiteralPosition):
            # Create a constant attribute
            create_attr_op = pdl_interp.CreateAttributeOp(position.value)
            self.builder.insert(create_attr_op)
            value = create_attr_op.attribute

        elif isinstance(position, TypePosition):
            assert parent_val is not None
            # Get type of value or attribute
            if parent_val.type == pdl.AttributeType():
                raise NotImplementedError(
                    "pdl_interp.get_attribute_type is not yet implemented"
                )
                get_type_op = pdl_interp.GetAttributeTypeOp(parent_val)
            else:
                get_type_op = pdl_interp.GetValueTypeOp(parent_val)
            self.builder.insert(get_type_op)
            value = get_type_op.result

        elif isinstance(position, TypeLiteralPosition):
            # Create a constant type or types
            raw_type_attr = position.value
            if isinstance(raw_type_attr, TypeAttribute):
                create_type_op = pdl_interp.CreateTypeOp(raw_type_attr)
                self.builder.insert(create_type_op)
                value = create_type_op.result
            else:
                # Assume it's an ArrayAttr of types
                assert isinstance(raw_type_attr, ArrayAttr)
                type_attr = cast(ArrayAttr[TypeAttribute], raw_type_attr)
                create_types_op = pdl_interp.CreateTypesOp(type_attr)
                self.builder.insert(create_types_op)
                value = create_types_op.result

        elif isinstance(position, ConstraintPosition):
            # The constraint op has already been created, find it in the map
            constraint_op = self.constraint_op_map.get(position.constraint)
            assert constraint_op is not None
            value = constraint_op.results[position.result_index]

        elif isinstance(position, UsersPosition):
            raise NotImplementedError("UsersPosition not implemented in lowering")
        elif isinstance(position, ForEachPosition):
            raise NotImplementedError("ForEachPosition not implemented in lowering")
        else:
            raise NotImplementedError(f"Unhandled position type {type(position)}")

        # Cache and return
        assert value is not None
        self.values[position] = value
        return value

    def generate_bool_node(self, node: BoolNode, block: Block, val: SSAValue) -> None:
        """Generate operations for a boolean predicate node"""

        question = node.question
        answer = node.answer
        region = block.parent
        assert region is not None, "Block must be in a region"

        # Handle getValue queries first for constraint questions
        args: list[SSAValue] = []
        if isinstance(question, EqualToQuestion):
            args = [self.get_value_at(block, question.other_position)]
        elif isinstance(question, ConstraintQuestion):
            for position in question.arg_positions:
                args.append(self.get_value_at(block, position))

        # Create success block
        success_block = Block()
        region.add_block(success_block)
        failure_block = self.failure_block_stack[-1]

        # Generate predicate check operation based on question type
        match question:
            case IsNotNullQuestion():
                check_op = pdl_interp.IsNotNullOp(val, success_block, failure_block)
            case OperationNameQuestion():
                assert isinstance(answer, StringAnswer)
                check_op = pdl_interp.CheckOperationNameOp(
                    answer.value, val, success_block, failure_block
                )
            case OperandCountQuestion() | OperandCountAtLeastQuestion():
                assert isinstance(answer, UnsignedAnswer)
                compare_at_least = isinstance(question, OperandCountAtLeastQuestion)
                check_op = pdl_interp.CheckOperandCountOp(
                    val, answer.value, success_block, failure_block, compare_at_least
                )
            case ResultCountQuestion() | ResultCountAtLeastQuestion():
                assert isinstance(answer, UnsignedAnswer)
                compare_at_least = isinstance(question, ResultCountAtLeastQuestion)
                check_op = pdl_interp.CheckResultCountOp(
                    val, answer.value, success_block, failure_block, compare_at_least
                )
            case EqualToQuestion():
                # Get the other value to compare with
                other_val = self.get_value_at(block, question.other_position)
                assert isinstance(answer, TrueAnswer)
                check_op = pdl_interp.AreEqualOp(
                    val, other_val, success_block, failure_block
                )
            case AttributeConstraintQuestion():
                assert isinstance(answer, AttributeAnswer)
                check_op = pdl_interp.CheckAttributeOp(
                    answer.value, val, success_block, failure_block
                )
            case TypeConstraintQuestion():
                assert isinstance(answer, TypeAnswer)
                if isinstance(val.type, pdl.RangeType):
                    # Check multiple types
                    raise NotImplementedError(
                        "pdl_interp.check_types is not yet implemented"
                    )
                    check_op = pdl_interp.CheckTypesOp(
                        val, answer.value, success_block, failure_block
                    )
                else:
                    # Check single type
                    assert isinstance(answer.value, TypeAttribute)
                    check_op = pdl_interp.CheckTypeOp(
                        answer.value, val, success_block, failure_block
                    )
            case ConstraintQuestion():
                check_op = pdl_interp.ApplyConstraintOp(
                    question.name,
                    args,
                    success_block,
                    failure_block,
                    is_negated=question.is_negated,
                    res_types=question.result_types,
                )
                # Store the constraint op for later result access
                self.constraint_op_map[question] = check_op
            case _:
                raise NotImplementedError(f"Unhandled question type {type(question)}")

        self.builder.insert_op(check_op, InsertPoint.at_end(block))

        # Generate matcher for success node
        if node.success_node:
            self.generate_matcher(node.success_node, region, success_block)

    def generate_switch_node(
        self, node: SwitchNode, block: Block, val: SSAValue
    ) -> None:
        """Generate operations for a switch node"""

        question = node.question
        region = block.parent
        assert region is not None, "Block must be in a region"
        default_dest = self.failure_block_stack[-1]

        # Handle at-least questions specially
        if isinstance(
            question, OperandCountAtLeastQuestion | ResultCountAtLeastQuestion
        ):
            # Sort children in reverse numerical order
            sorted_children = sorted(
                node.children.items(),
                key=lambda x: cast(UnsignedAnswer, x[0]).value,
                reverse=True,
            )

            # Push temporary entry to failure block stack
            self.failure_block_stack.append(default_dest)

            for answer, child_node in sorted_children:
                if child_node:
                    success_block = self.generate_matcher(child_node, region)
                    current_check_block = Block()
                    region.add_block(current_check_block)

                    self.builder.insertion_point = InsertPoint.at_end(
                        current_check_block
                    )
                    assert isinstance(answer, UnsignedAnswer)
                    if isinstance(question, OperandCountAtLeastQuestion):
                        check_op = pdl_interp.CheckOperandCountOp(
                            val,
                            answer.value,
                            success_block,
                            default_dest,
                            True,
                        )
                    else:
                        check_op = pdl_interp.CheckResultCountOp(
                            val,
                            answer.value,
                            success_block,
                            default_dest,
                            True,
                        )
                    self.builder.insert(check_op)

                    # Update failure block stack for next child matcher
                    self.failure_block_stack[-1] = current_check_block

            # Pop the temporary entry from failure block stack
            first_predicate_block = self.failure_block_stack.pop()

            # Move ops from the first check block into the main block
            for op in list(first_predicate_block.ops):
                op.detach()
                block.add_op(op)
            assert first_predicate_block.parent is not None
            first_predicate_block.parent.detach_block(first_predicate_block)
            first_predicate_block.erase()

            return

        # Generate child blocks and collect case values
        case_blocks: list[Block] = []
        case_values: list[Answer] = []

        for answer, child_node in node.children.items():
            if child_node:
                child_block = self.generate_matcher(child_node, region)
                case_blocks.append(child_block)
                case_values.append(answer)

        # Position builder at end of current block
        self.builder.insertion_point = InsertPoint.at_end(block)

        # Create switch operation based on question type
        match question:
            case OperationNameQuestion():
                # Extract string values from StringAnswer objects
                switch_values = [cast(StringAnswer, ans).value for ans in case_values]
                switch_attr = ArrayAttr([StringAttr(v) for v in switch_values])
                switch_op = pdl_interp.SwitchOperationNameOp(
                    switch_attr, val, default_dest, case_blocks
                )
            case OperandCountQuestion():
                # Extract integer values from UnsignedAnswer objects
                switch_values = [cast(UnsignedAnswer, ans).value for ans in case_values]
                switch_attr = ArrayAttr([IntegerAttr(v, 32) for v in switch_values])
                raise NotImplementedError(
                    "pdl_interp.switch_operand_count is not yet implemented"
                )
                switch_op = pdl_interp.SwitchOperandCountOp(
                    switch_attr, val, default_dest, case_blocks
                )
            case ResultCountQuestion():
                # Extract integer values from UnsignedAnswer objects
                switch_values = [cast(UnsignedAnswer, ans).value for ans in case_values]
                switch_attr = ArrayAttr([IntegerAttr(v, 32) for v in switch_values])
                raise NotImplementedError(
                    "pdl_interp.switch_result_count is not yet implemented"
                )
                switch_op = pdl_interp.SwitchResultCountOp(
                    switch_attr, val, default_dest, case_blocks
                )
            case TypeConstraintQuestion():
                # Extract type attributes from TypeAnswer objects
                switch_values = [cast(TypeAnswer, ans).value for ans in case_values]
                raise NotImplementedError(
                    "pdl_interp.switch_types is not yet implemented"
                )
                if isinstance(val.type, pdl.RangeType):
                    switch_attr = ArrayAttr(switch_values)

                    switch_op = pdl_interp.SwitchTypesOp(
                        switch_attr, val, default_dest, case_blocks
                    )
                else:
                    switch_attr = ArrayAttr(switch_values)
                    switch_op = pdl_interp.SwitchTypeOp(
                        switch_attr, val, default_dest, case_blocks
                    )
            case AttributeConstraintQuestion():
                # Extract attribute values from AttributeAnswer objects
                switch_values = [
                    cast(AttributeAnswer, ans).value for ans in case_values
                ]
                switch_attr = ArrayAttr(switch_values)
                switch_op = pdl_interp.SwitchAttributeOp(
                    val, switch_attr, default_dest, case_blocks
                )
            case _:
                raise NotImplementedError(f"Unhandled question type {type(question)}")

        self.builder.insert(switch_op)

    def _generate_rewriter_for_apply_native_rewrite(
        self,
        op: pdl.ApplyNativeRewriteOp,
        rewrite_values: dict[SSAValue, SSAValue],
        map_rewrite_value: Callable[[SSAValue], SSAValue],
    ):
        arguments = [map_rewrite_value(arg) for arg in op.args]
        result_types = [res.type for res in op.res]
        raise NotImplementedError("pdl_interp.apply_rewrite is not yet implemented")
        interp_op = pdl_interp.ApplyRewriteOp(
            arguments, name=op.constraint_name, result_types=result_types
        )
        self.rewriter_builder.insert(interp_op)
        for old_res, new_res in zip(op.results, interp_op.results, strict=True):
            rewrite_values[old_res] = new_res

    def _generate_rewriter_for_attribute(
        self,
        op: pdl.AttributeOp,
        rewrite_values: dict[SSAValue, SSAValue],
        map_rewrite_value: Callable[[SSAValue], SSAValue],
    ):
        if op.value is not None:
            new_attr_op = pdl_interp.CreateAttributeOp(op.value)
            self.rewriter_builder.insert(new_attr_op)
            rewrite_values[op.output] = new_attr_op.attribute



import marimo

__generated_with = "0.15.2"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo
    return (mo,)


@app.cell
def _():
    from xdsl.utils import marimo as xmo
    return (xmo,)


@app.cell(hide_code=True)
def _():
    from xdsl.dialects import arith, builtin, pdl, func
    from xdsl.context import Context
    from xdsl.parser import Parser
    from xdsl.transforms.apply_pdl import ApplyPDLPass
    from xdsl.transforms.dead_code_elimination import dce
    from xdsl.rewriter import Rewriter
    from xdsl.builder import InsertPoint

    ctx = Context()
    ctx.load_dialect(builtin.Builtin)
    ctx.load_dialect(arith.Arith)
    ctx.load_dialect(pdl.PDL)
    ctx.load_dialect(func.Func)
    return (
        ApplyPDLPass,
        InsertPoint,
        Parser,
        Rewriter,
        arith,
        builtin,
        ctx,
        dce,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md("""# Pattern Description Language (PDL)""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    One of the most frequent kinds of transformations on intermediate representations in compilers are local, "peephole", rewrites, which transform an operation by inspecting its local context only.
    An example of this might be rewriting `arith.add` into `llvm.add`, which can be done just by inspecting the type of the operation, or local optimisations such as rewriting an `arith.add` with `0` to the other operand.
    In both MLIR and xDSL these can be written using the native language API (C++ or Python), a flexible and productive approach with two important flaws: reasoning about these rewrites requires reasoning about the semantics of the host language, which is famously difficult for both C++ and Python, and generating them is much less convenient than generating MLIR IR directly.
    The `pdl` dialect addresses both of these issues.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## PDL Patterns""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""First let's look at the pattern that rewrites `first` to `second`:""")
    return


@app.cell(hide_code=True)
def _(Parser, builtin, ctx, xmo):
    first_text = """
    func.func @first(%x : i32) -> i32 {
      %c0 = arith.constant 0 : i32
      %y = arith.muli %x, %c0 : i32
      func.return %y : i32
    }
    """

    second_text = """
    func.func @second(%x : i32) -> i32 {
      %c0 = arith.constant 0 : i32
      func.return %c0 : i32
    }
    """

    first_op = Parser(ctx, first_text).parse_op()
    second_op = Parser(ctx, second_text).parse_op()

    xmo.module_html(builtin.ModuleOp([first_op.clone(), second_op.clone()]))
    return first_op, second_op


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    Note that these are equivalent, as `x * 0 = 0` for all `x`.

    The following pattern implements the rewrite for 32-bit integers:
    """
    )
    return


@app.cell(hide_code=True)
def _(Parser, ctx, xmo):
    times_zero_text = """
    pdl.pattern @x_times_zero : benefit(2) {
      %t = pdl.type
      %x = pdl.operand
      %c0_attr = pdl.attribute = 0 : i32
      %c0_op = pdl.operation "arith.constant" {"value" = %c0_attr} -> (%t : !pdl.type)
      %c0_res = pdl.result 0 of %c0_op
      %x_times_zero_op = pdl.operation "arith.muli" (%x, %c0_res : !pdl.value, !pdl.value) -> (%t : !pdl.type)
      pdl.rewrite %x_times_zero_op {
        pdl.replace %x_times_zero_op with (%c0_res : !pdl.value)
      }
    }
    """

    times_zero_op = Parser(ctx, times_zero_text).parse_op()

    xmo.module_html(times_zero_op)
    return (times_zero_op,)


@app.cell(hide_code=True)
def _(ApplyPDLPass, builtin, ctx, first_op, second_op, times_zero_op):
    def test_rewrite():
        input_copy = first_op.clone()
        input_copy.sym_name = builtin.StringAttr("second")
        pattern_copy = times_zero_op.clone()
        module = builtin.ModuleOp([input_copy, pattern_copy])
        ApplyPDLPass().apply(ctx, module)
        assert str(input_copy) == str(second_op)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    Let's look at the pattern in more detail.

    The [`pdl.pattern` operation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlpattern-pdlpatternop) has an optional name, a "benefit" field, and a body.

    ```
    pdl.pattern : benefit(2) {
      ...
    }
    ```

    The non-negative benefit is used in the case where multiple patterns match the same program.

    The body consists of two parts, a declarative matching region terminated with a [`pdl.rewrite` operation](https://mlir.llvm.org/docs/Dialects/PDLOps/#pdlrewrite-pdlrewriteop) containing an imperative rewrite region:

    ``` C
    // Match old IR
    pdl.rewrite %root {
      // Create new IR
    }
    ```

    The rewrite takes a number of arguments which correspond to the operation being rewritten.
    In practice, there is almost always one operation being matched on.

    Here's the matching region with comments:

    ```
    // 1. The type that we're operating on
    %t = pdl.type
    // 2. The lhs operand to the operation
    %x = pdl.operand
    // 3. The zero attribute of our target type
    %c0_attr = pdl.attribute = 0 : i32
    // 4. A specification of the constant operation we want to match on
    %c0_op = pdl.operation "arith.constant" {"value" = %c0_attr} -> (%t : !pdl.type)
    // 5. The result value of the constant operation
    %c0_res = pdl.result 0 of %c0_op
    // 6. Finally, the operation that we want to rewrite
    %x_times_zero_op = pdl.operation "arith.muli" (%x, %c0_res : !pdl.value, !pdl.value) -> (%t : !pdl.type)
    ```

    Note that in this region, we don't have to provide all of the components of the matched operation, instead we only have to provide the constraints that we care about.

    This example's rewrite operation is quite simple, replacing the matched operation with the zero operand:

    ```
    pdl.rewrite %x_times_zero_op {
      pdl.replace %x_times_zero_op with (%c0_res : !pdl.value)
    }
    ```

    Please refer to the [dialect reference](https://mlir.llvm.org/docs/Dialects/PDLOps/) for the full list of operations and types.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo, write_check):
    mo.md(
        rf"""
    ## Exercises {write_check}

    In this notebook, you'll write patterns in PDL to transform this input function:
    """
    )
    return


@app.cell(hide_code=True)
def _(arith, builtin):
    from xdsl.frontend.pyast.context import PyASTContext
    from xdsl.frontend.pyast.utils.exceptions import CodeGenerationException

    # Set up the AST parsing context
    pyast_ctx = PyASTContext()
    pyast_ctx.register_type(float, builtin.f64)
    pyast_ctx.register_function(float.__add__, arith.AddfOp)
    pyast_ctx.register_function(float.__sub__, arith.SubfOp)
    pyast_ctx.register_function(float.__mul__, arith.MulfOp)
    pyast_ctx.register_function(float.__truediv__, arith.DivfOp)
    return (pyast_ctx,)


@app.cell
def _(mo, pyast_ctx):
    @pyast_ctx.parse_program
    def main(a: float, b: float, c: float) -> float:
        return (c + (a - a)) / (b / b)

    import inspect
    lines = inspect.getsource(main)
    mo.md(f"```python\n{lines}```")
    return (main,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""As you can see, it just returns `c`, we'll want to rewrite our program to return `c` directly, without the useless computations.""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Here is the corresponding IR:""")
    return


@app.cell(hide_code=True)
def _(main, xmo):
    xmo.module_html(main.module.body.ops.first)
    return


@app.cell(hide_code=True)
def _(mo):
    x_minus_x_text = """\
    pdl.pattern @x_minus_x : benefit(2) {
      %t = pdl.type
      %x = pdl.operand

      // Add op to match here

      // Uncomment these lines:
      // pdl.rewrite %x_minus_x {
      //   %c0_attr = pdl.attribute = 0.0 : f64
      //   %c0_op = pdl.operation "arith.constant" {"value" = %c0_attr} -> (%t : !pdl.type)
      //   %c0_res = pdl.result 0 of %c0_op
      //   pdl.replace %x_minus_x with (%c0_res : !pdl.value)
      // }
    }"""

    x_minus_x_text_area = mo.ui.code_editor(x_minus_x_text, language="javascript")
    return (x_minus_x_text_area,)


@app.cell(hide_code=True)
def _(info_text, mo, x_minus_x_text_area):
    mo.vstack((
        mo.md("### Exercise 1\nFill out the matching region below to implement `x - x -> 0`:"),
        x_minus_x_text_area,
        mo.md(info_text))
    )
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    x_minus_x_text_solution = """\
    pdl.pattern @x_minus_x : benefit(2) {
      %t = pdl.type
      %x = pdl.operand
      %x_minus_x = pdl.operation "arith.subf" (%x, %x : !pdl.value, !pdl.value) -> (%t : !pdl.type)
      pdl.rewrite %x_minus_x {
        %c0_attr = pdl.attribute = 0.0 : f64
        %c0_op = pdl.operation "arith.constant" {"value" = %c0_attr} -> (%t : !pdl.type)
        %c0_res = pdl.result 0 of %c0_op
        pdl.replace %x_minus_x with (%c0_res : !pdl.value)
      }
    }"""

    mo.accordion({"Solution": xmo.module_html(x_minus_x_text_solution)})
    return (x_minus_x_text_solution,)


@app.cell(hide_code=True)
def _(mo):
    x_plus_zero_text = """\
    pdl.pattern @x_plus_zero : benefit(2) {
      %t = pdl.type
      %x = pdl.operand

      // Implement matching ops here
      // %c0_attr =
      // %c0_op =
      // %c0_res = pdl.result 0 of %c0_op
      // %x_plus_zero_op =

      // Uncomment these lines:
      // pdl.rewrite %x_plus_zero_op {
      //   pdl.replace %x_plus_zero_op with (%x : !pdl.value)
      // }
    }"""
    x_plus_zero_text_area = mo.ui.code_editor(x_plus_zero_text, language="javascript")
    return (x_plus_zero_text_area,)


@app.cell(hide_code=True)
def _(info_text, mo, x_plus_zero_text_area):
    mo.vstack((
        mo.md("### Exercise 2\nFill out the matching region below to implement `x + 0 -> x`:"),
        x_plus_zero_text_area,
        mo.md(info_text))
    )
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    x_plus_zero_text_solution = """\
    pdl.pattern @x_plus_zero : benefit(2) {
      %t = pdl.type
      %x = pdl.operand
      %c0_attr = pdl.attribute = 0.0 : f64
      %c0_op = pdl.operation "arith.constant" {"value" = %c0_attr} -> (%t : !pdl.type)
      %c0_res = pdl.result 0 of %c0_op
      %x_plus_zero_op = pdl.operation "arith.addf" (%x, %c0_res : !pdl.value, !pdl.value) -> (%t : !pdl.type)
      pdl.rewrite %x_plus_zero_op {
        pdl.replace %x_plus_zero_op with (%x : !pdl.value)
      }
    }"""

    mo.accordion({"Solution": xmo.module_html(x_plus_zero_text_solution)})
    return (x_plus_zero_text_solution,)


@app.cell(hide_code=True)
def _(mo):
    x_div_x_text = """\
    pdl.pattern @x_div_x : benefit(2) {
      %t = pdl.type
      %x = pdl.operand
      %x_div_x_op = pdl.operation "arith.divf" (%x, %x : !pdl.value, !pdl.value) -> (%t : !pdl.type)
      pdl.rewrite %x_div_x_op {
        // Implement rewrite region here
      }
    }"""
    x_div_x_text_area = mo.ui.code_editor(x_div_x_text, language="javascript")
    return (x_div_x_text_area,)


@app.cell(hide_code=True)
def _(info_text, mo, x_div_x_text_area):
    mo.vstack((
        mo.md("### Exercise 3\nFill out the matching region to implement `x / x -> 1`:"),
        x_div_x_text_area,
        mo.md(info_text))
    )
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    x_div_x_text_solution = """\
    pdl.pattern @x_div_x : benefit(2) {
      %t = pdl.type
      %x = pdl.operand
      %x_div_x_op = pdl.operation "arith.divf" (%x, %x : !pdl.value, !pdl.value) -> (%t : !pdl.type)
      pdl.rewrite %x_div_x_op {
        %c1_attr = pdl.attribute = 1.0 : f64
        %c1_op = pdl.operation "arith.constant" {"value" = %c1_attr} -> (%t : !pdl.type)
        %c1_res = pdl.result 0 of %c1_op
        pdl.replace %x_div_x_op with (%c1_res : !pdl.value)
      }
    }"""

    mo.accordion({"Solution": xmo.module_html(x_div_x_text_solution)})
    return (x_div_x_text_solution,)


@app.cell(hide_code=True)
def _(mo):
    x_div_one_text = """\
    pdl.pattern @x_div_one : benefit(2) {
      %t = pdl.type
      %x = pdl.operand

      // Implement matching ops here
      // %c1_attr =
      // %c1_op =
      // %c1_res = pdl.result 0 of %c1_op
      // %x_div_one_op =

      // Uncomment these lines:
      // pdl.rewrite %x_div_one_op {
      //   pdl.replace %x_div_one_op with (%x : !pdl.value)
      // }
    }"""
    x_div_one_text_area = mo.ui.code_editor(x_div_one_text, language="javascript")
    return (x_div_one_text_area,)


@app.cell(hide_code=True)
def _(info_text, mo, x_div_one_text_area):
    mo.vstack((
        mo.md("### Exercise 4\nFill out the matching region below to implement `x / 1 -> x`:"),
        x_div_one_text_area,
        mo.md(info_text))
    )
    return


@app.cell(hide_code=True)
def _(mo, xmo):
    x_div_one_text_solution = """\
    pdl.pattern @x_div_one : benefit(2) {
      %t = pdl.type
      %x = pdl.operand
      %c1_attr = pdl.attribute = 1.0 : f64
      %c1_op = pdl.operation "arith.constant" {"value" = %c1_attr} -> (%t : !pdl.type)
      %c1_res = pdl.result 0 of %c1_op
      %x_div_one_op = pdl.operation "arith.divf" (%x, %c1_res : !pdl.value, !pdl.value) -> (%t : !pdl.type)
      pdl.rewrite %x_div_one_op {
        pdl.replace %x_div_one_op with (%x : !pdl.value)
      }
    }"""

    mo.accordion({"Solution": xmo.module_html(x_div_one_text_solution)})
    return (x_div_one_text_solution,)


@app.cell
def _():
    expected_text = """\
    func.func @main(%a : f64, %b : f64, %c : f64) -> f64 {
      func.return %c : f64
    }"""
    return (expected_text,)


@app.cell
def _(
    ApplyPDLPass,
    InsertPoint,
    Parser,
    Rewriter,
    ctx,
    dce,
    expected_text,
    main,
    x_div_one_text_area,
    x_div_x_text_area,
    x_minus_x_text_area,
    x_plus_zero_text_area,
):
    _error_text = ""
    _results_text = ""
    try:
        _patterns_text = x_minus_x_text_area.value + x_plus_zero_text_area.value + x_div_x_text_area.value + x_div_one_text_area.value
        _module = Parser(ctx, _patterns_text).parse_module()
        _cloned_func = main.module.body.ops.first.clone()
        Rewriter.insert_op(_cloned_func, InsertPoint.at_start(_module.body.block))
        ApplyPDLPass().apply(ctx, _module)
        dce(_module)
        _results_text = str(_cloned_func)
        write_check = "✅" if _results_text == expected_text else "❌"
        if _results_text == expected_text:
            check_text = f"{write_check} Applying all patterns yields expected function."
        else:
            check_text = f"{write_check} Applying all patterns does not yield expected function."
    except Exception as e:
        _error_text = str(e)
    if _error_text:
        info_text = f"""
    /// attention | Error:

    ```
    {_error_text}
    ```
    """
    else:
        info_text = f"""
    Here are the outputs when running the function:

    ```
    {_results_text}
    ```

    {check_text}
    """
    return info_text, write_check


@app.cell(hide_code=True)
def _(
    ApplyPDLPass,
    InsertPoint,
    Parser,
    Rewriter,
    ctx,
    dce,
    expected_text,
    main,
    x_div_one_text_solution,
    x_div_x_text_solution,
    x_minus_x_text_solution,
    x_plus_zero_text_solution,
):
    def test_solutions():
        _solutions_text = x_minus_x_text_solution + x_plus_zero_text_solution + x_div_x_text_solution + x_div_one_text_solution
        _module = Parser(ctx, _solutions_text).parse_module()
        _cloned_func = main.module.body.ops.first.clone()
        Rewriter.insert_op(_cloned_func, InsertPoint.at_start(_module.body.block))
        ApplyPDLPass().apply(ctx, _module)
        dce(_module)
        _results_text = str(_cloned_func)
        assert _results_text == expected_text
    return


if __name__ == "__main__":
    app.run()



from xdsl.backend.assembly_printer import reg
from xdsl.dialects.arm.registers import IntRegisterType
from xdsl.utils.test_value import create_ssa_value


def test_assembly_arg_str_ARMRegister():
    x0 = IntRegisterType.from_name("x0")
    x0_val = create_ssa_value(x0)
    assert reg(x0_val) == "x0"



import marimo

__generated_with = "0.15.3"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo
    return (mo,)


@app.cell(hide_code=True)
def _():
    from xdsl.utils import marimo as xmo
    return (xmo,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
    A small test-page to test mkdocs-marimo.

    ## Input
    """
    )
    return


@app.cell
def _(mo):
    expr_str = mo.ui.code_editor(value = "let c = 1 + 2;\nc + 3", language = "rust", debounce=False)
    expr_str
    return (expr_str,)


@app.cell(hide_code=True)
def _(mo):
    mo.md("""## Output""")
    return


@app.cell
def _(mo):
    get_state, set_state = mo.state("")
    return get_state, set_state


@app.cell
def _(expr_str, get_state, mo, set_state, xmo):
    from xdsl.frontend.listlang.main import ParseError, parse_program
    from xdsl.dialects import builtin
    from xdsl.builder import Builder, InsertPoint

    def to_mlir(code: str) -> builtin.ModuleOp:
        module = builtin.ModuleOp([])
        builder = Builder(InsertPoint.at_start(module.body.block))
        parse_program(code, builder)
        return module

    try:
        def convert(code) -> str:
            output = to_mlir(code.value)
            return output

        res = convert(expr_str)
        set_state(res)
    except ParseError:
        res = get_state()

    res_str = xmo.module_md(res).text.replace("\n", "<br>")

    mo.md(f"{res_str}")
    return (res,)


@app.cell
def _(mo):
    mo.md(r"""# Module post-modification""")
    return


@app.cell
def _():
    pass_list = ["constant-fold-interp","canonicalize"]
    return (pass_list,)


@app.cell
def _(mo, pass_list):
    slider_choices = ["IR before a pass was executed"] + ["IR after " + name for name in pass_list]
    slider = mo.ui.slider(start=0, stop=len(slider_choices) - 1, label="Slider", value=0)
    return slider, slider_choices


@app.cell
def _(mo, slider, slider_choices):
    stack = mo.hstack([slider, mo.md(slider_choices[slider.value])])
    stack
    return


@app.cell
def _(mo, pass_list, res, slider, xmo):
    from xdsl.passes import PassPipeline
    from xdsl.transforms import get_all_passes
    from xdsl.dialects import get_all_dialects
    from xdsl.context import Context

    module = res.clone()


    module_list = [module.clone()]

    def callback(pass1, module, pass2):
        module_list.append(module.clone())

    pipeline = PassPipeline.parse_spec(get_all_passes(), ",".join(pass_list), callback)

    ctx = Context()
    for name, func in get_all_dialects().items():
        ctx.register_dialect(name, func)


    pipeline.apply(ctx, module)
    module_list.append(module.clone())

    res_str_0 = xmo.module_md(module_list[slider.value]).text.replace("\n", "<br>")

    mo.md(res_str_0)
    return


if __name__ == "__main__":
    app.run()



from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Operation
from xdsl.irdl import IRDLOperation, irdl_op_definition, traits_def
from xdsl.pattern_rewriter import PatternRewriter, RewritePattern
from xdsl.traits import HasCanonicalizationPatternsTrait
from xdsl.transforms.canonicalize import CanonicalizePass


class Pattern(RewritePattern):
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter, /):
        rewriter.erase_matched_op()


class HasCanonicalizationPatternsTrait1(HasCanonicalizationPatternsTrait):
    @classmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        return ()


class HasCanonicalizationPatternsTrait2(HasCanonicalizationPatternsTrait):
    @classmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        return (Pattern(),)


@irdl_op_definition
class TestOp(IRDLOperation):
    name = "test.op"

    traits = traits_def(
        HasCanonicalizationPatternsTrait1(), HasCanonicalizationPatternsTrait2()
    )


def test_multiple_traits():
    """
    Check that operations with multiple canonicalization patterns traits get patterns
    from all the traits and not just the first one.
    """
    op = TestOp()
    module = ModuleOp([op])
    ctx = Context()
    ctx.load_op(ModuleOp)
    ctx.load_op(TestOp)

    CanonicalizePass().apply(ctx, module)

    assert not module.body.block.ops



import re
from io import StringIO

import pytest

from xdsl.context import Context
from xdsl.dialects.builtin import IndexType, IntegerType, i32
from xdsl.dialects.utils import (
    get_dynamic_index_list,
    parse_dynamic_index_list_with_types,
    parse_dynamic_index_list_without_types,
    parse_dynamic_index_with_type,
    parse_dynamic_index_without_type,
    parse_empty_dimension_list_directive,
    print_dynamic_index_list,
    print_empty_dimension_list_directive,
    split_dynamic_index_list,
)
from xdsl.dialects.utils.dynamic_index_list import verify_dynamic_index_list
from xdsl.ir import Dialect, SSAValue
from xdsl.parser import Parser, UnresolvedOperand
from xdsl.printer import Printer
from xdsl.utils.exceptions import ParseError, VerifyException
from xdsl.utils.test_value import create_ssa_value

ctx = Context()
index = IndexType()

DYNAMIC_INDEX = -42


def test_split_dynamic_index_list():
    # Test case 1: Only integers
    values = [1, 2, 3]
    static, dynamic = split_dynamic_index_list(values, DYNAMIC_INDEX)
    assert static == [1, 2, 3]
    assert dynamic == []

    # Test case 2: Mix of integers and SSA values
    val1 = create_ssa_value(IndexType())
    val2 = create_ssa_value(IndexType())
    values = [1, 2, val1, 4, val2]
    static, dynamic = split_dynamic_index_list(values, DYNAMIC_INDEX)
    assert static == [1, 2, DYNAMIC_INDEX, 4, DYNAMIC_INDEX]
    assert dynamic == [val1, val2]

    # Test case 3: All SSA values
    values = [val1, val2]
    static, dynamic = split_dynamic_index_list(values, DYNAMIC_INDEX)
    assert static == [DYNAMIC_INDEX, DYNAMIC_INDEX]
    assert dynamic == [val1, val2]

    # Test case 4: Empty list
    static, dynamic = split_dynamic_index_list([], DYNAMIC_INDEX)
    assert static == []
    assert dynamic == []


def test_get_dynamic_index_list():
    # Test case 1: Only integers
    static_values = [1, 2, 3]
    result = get_dynamic_index_list(static_values, [], DYNAMIC_INDEX)
    assert result == [1, 2, 3]

    # Test case 2: Mix of integers and SSA values
    val1 = create_ssa_value(IndexType())
    val2 = create_ssa_value(IndexType())
    static_values = [1, 2, DYNAMIC_INDEX, 4, DYNAMIC_INDEX]
    dynamic_values = [val1, val2]
    result = get_dynamic_index_list(static_values, dynamic_values, DYNAMIC_INDEX)
    assert result == [1, 2, val1, 4, val2]

    # Test case 3: All SSA values
    static_values = [DYNAMIC_INDEX, DYNAMIC_INDEX]
    dynamic_values = [val1, val2]
    result = get_dynamic_index_list(static_values, dynamic_values, DYNAMIC_INDEX)
    assert result == [val1, val2]


def test_verify_dynamic_index_list():
    # Test case 1: Valid input
    static_values = [1, 2, DYNAMIC_INDEX]
    dynamic_values = [create_ssa_value(IndexType())]
    verify_dynamic_index_list(static_values, dynamic_values, DYNAMIC_INDEX)

    # Test case 2: Invalid input (mismatched lengths)
    static_values = [1, 2, DYNAMIC_INDEX]
    with pytest.raises(
        VerifyException,
        match=re.escape(
            "The number of dynamic positions passed as values (0) does not match "
            "the number of dynamic position markers (1)."
        ),
    ):
        verify_dynamic_index_list(static_values, [], DYNAMIC_INDEX)


def test_print_dynamic_index_list():
    # Test case 1: Only integers
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(printer, DYNAMIC_INDEX, [], [1, 2, 3])
    assert stream.getvalue() == "[1, 2, 3]"

    # Test case 2: Mix of integers and SSA values
    stream = StringIO()
    printer = Printer(stream)
    values = [create_ssa_value(IndexType()), create_ssa_value(IndexType())]
    print_dynamic_index_list(
        printer,
        DYNAMIC_INDEX,
        values,
        [1, DYNAMIC_INDEX, 3, DYNAMIC_INDEX],
    )
    assert stream.getvalue() == "[1, %0, 3, %1]"

    # Test case 3: With value types
    stream = StringIO()
    printer = Printer(stream)
    values = [create_ssa_value(IndexType()), create_ssa_value(IntegerType(32))]
    value_types = (IndexType(), IntegerType(32))
    print_dynamic_index_list(
        printer,
        DYNAMIC_INDEX,
        values,
        [DYNAMIC_INDEX, 2, DYNAMIC_INDEX],
        value_types,
    )
    assert stream.getvalue() == "[%0 : index, 2, %1 : i32]"

    # Test case 4: Custom delimiter
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(
        printer,
        DYNAMIC_INDEX,
        [],
        [1, 2, 3],
        delimiter=Parser.Delimiter.PAREN,
    )
    assert stream.getvalue() == "(1, 2, 3)"

    # Test case 5: Empty list
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(printer, DYNAMIC_INDEX, [], [])
    assert stream.getvalue() == "[]"


@pytest.mark.parametrize(
    "delimiter,expected",
    [
        (Parser.Delimiter.SQUARE, "[1, 2, 3]"),
        (Parser.Delimiter.PAREN, "(1, 2, 3)"),
        (Parser.Delimiter.ANGLE, "<1, 2, 3>"),
        (Parser.Delimiter.BRACES, "{1, 2, 3}"),
    ],
)
def test_print_dynamic_index_list_delimiters(
    delimiter: Parser.Delimiter, expected: str
):
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(printer, DYNAMIC_INDEX, [], [1, 2, 3], delimiter=delimiter)
    assert stream.getvalue() == expected


def test_parse_dynamic_index_with_type():
    parser = Parser(ctx, "%0 : i32")
    result = parse_dynamic_index_with_type(parser)
    assert isinstance(result, SSAValue)

    parser = Parser(ctx, "42")
    result = parse_dynamic_index_with_type(parser)
    assert isinstance(result, int)
    assert result == 42


def test_parse_dynamic_index_without_type():
    parser = Parser(ctx, "%0")
    result = parse_dynamic_index_without_type(parser)
    assert isinstance(result, UnresolvedOperand)

    parser = Parser(ctx, "42")
    result = parse_dynamic_index_without_type(parser)
    assert isinstance(result, int)
    assert result == 42


def test_parse_dynamic_index_list_with_types():
    dynamic_index = -42
    test_values = (create_ssa_value(i32), create_ssa_value(index))
    parser = Parser(ctx, "[%0 : i32, 42, %1 : index]")
    parser.ssa_values["0"] = (test_values[0],)
    parser.ssa_values["1"] = (test_values[1],)
    values, indices = parse_dynamic_index_list_with_types(
        parser, dynamic_index=dynamic_index
    )
    assert len(values) == 2
    assert values[0] is test_values[0]
    assert values[1] is test_values[1]
    assert tuple(indices) == (dynamic_index, 42, dynamic_index)


def test_parse_dynamic_index_list_without_types():
    dynamic_index = -42
    parser = Parser(ctx, "[%0, 42, %1]")
    values, indices = parse_dynamic_index_list_without_types(
        parser, dynamic_index=dynamic_index
    )

    assert len(values) == 2
    assert isinstance(values[0], UnresolvedOperand)
    assert isinstance(values[1], UnresolvedOperand)

    assert tuple(indices) == (dynamic_index, 42, dynamic_index)


def test_parse_dynamic_index_list_with_custom_delimiter():
    dynamic_index = -42
    test_values = (create_ssa_value(i32), create_ssa_value(index))
    parser = Parser(ctx, "(%0 : i32, 42, %1 : index)")
    parser.ssa_values["0"] = (test_values[0],)
    parser.ssa_values["1"] = (test_values[1],)
    values, indices = parse_dynamic_index_list_with_types(
        parser, dynamic_index=dynamic_index, delimiter=Parser.Delimiter.PAREN
    )
    assert len(values) == 2
    assert values[0] is test_values[0]
    assert values[1] is test_values[1]
    assert tuple(indices) == (dynamic_index, 42, dynamic_index)


@pytest.mark.parametrize(
    "string,expected",
    [
        ("[]", True),
        ("2x3x4", False),
        ("[2, 3, 4]", None),
    ],
)
def test_parse_empty_dimension_list(string: str, expected: bool | None):
    parser = Parser(ctx, string)

    if expected is not None:
        assert parse_empty_dimension_list_directive(parser) == expected
        return

    with pytest.raises(
        ParseError,
        match="']' expected",
    ):
        parse_empty_dimension_list_directive(parser)


def test_print_empty_dimension_list():
    stream = StringIO()
    printer = Printer(stream)
    print_empty_dimension_list_directive(printer)
    assert stream.getvalue() == "[]"


@pytest.mark.parametrize(
    "name,expected_1,expected_2",
    [
        ("dialect.op_name", "dialect", "op_name"),
        ("dialect.op.name", "dialect", "op.name"),
    ],
)
def test_split_name(name: str, expected_1: str, expected_2: str):
    result_1, result_2 = Dialect.split_name(name)
    assert result_1 == expected_1
    assert result_2 == expected_2


def test_split_name_failure():
    with pytest.raises(ValueError, match="Invalid operation or attribute name test."):
        Dialect.split_name("test")



from xdsl.context import Context
from xdsl.dialects import builtin, eqsat
from xdsl.ir import OpResult
from xdsl.passes import ModulePass
from xdsl.rewriter import Rewriter


def eqsat_extract(module_op: builtin.ModuleOp):
    eclass_ops = [op for op in module_op.walk() if isinstance(op, eqsat.AnyEClassOp)]

    while eclass_ops:
        op = eclass_ops.pop()
        if not op.result.uses:
            # Erase all operands and uses of operands
            ops_to_erase = [op] + [
                operand.owner
                for operand in op.operands
                if isinstance(operand, OpResult)
            ]
        elif (min_cost_index := op.min_cost_index) is not None:
            # Replace eclass result by operand
            operand = op.operands[min_cost_index.data]
            op.result.replace_by_if(operand, lambda use: use.operation is not op)
            # Erase eclass and all operand ops excluding min cost one
            ops_to_erase = [op] + [
                operand.owner
                for i, operand in enumerate(op.operands)
                if i != min_cost_index.data and isinstance(operand, OpResult)
            ]
            # Delete cost
            if (
                isinstance(operand, OpResult)
                and eqsat.EQSAT_COST_LABEL in operand.op.attributes
            ):
                del operand.op.attributes[eqsat.EQSAT_COST_LABEL]

        else:
            # Don't touch this eclass or its operands
            ops_to_erase = ()

        for op in ops_to_erase:
            Rewriter.erase_op(op)

    assert not eclass_ops


class EqsatExtractPass(ModulePass):
    """
    Extracts the subprogram with the lowest cost, as specified by the `min_cost_index`
    """

    name = "eqsat-extract"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        eqsat_extract(op)



async def _():
    # Get the current notebook URL, drop the 'blob' URL components that seem to be added,
    # and add the buildnumber that a makethedocs PR build seems to add. This allows to load
    # the wheel both locally and when deployed to makethedocs.
    async def import_xdsl():
        import re
        from urllib.parse import urlparse

        import micropip
        from marimo import notebook_location

        url = str(notebook_location()).replace("blob:", "")
        print(f"DEBUG: notebook url (full): {url}")

        url_parsed = urlparse(url)
        scheme = url_parsed.scheme
        netloc = url_parsed.netloc
        path = url_parsed.path

        print(f"DEBUG: notebook url (parsed): {url_parsed}")

        url = re.sub(
            "([^/])/([a-f0-9-]+-[a-f0-9-]+-[a-f0-9-]+-[a-f0-9-]+)", "\\1/", url, count=1
        )
        buildnumber = re.sub(".*--([0-9+]+).*", "\\1", url, count=1)

        new_url = scheme + "://" + netloc

        if buildnumber != url:
            new_url = new_url + "/" + buildnumber + "/"
        elif netloc == "xdsl.readthedocs.io":
            new_url = new_url + "/" + (path.split("/")[1])

        print(f"DEBUG: notebook url (trimmed): {new_url}")

        await micropip.install("xdsl @ " + new_url + "/xdsl-0.0.0-py3-none-any.whl")

    await import_xdsl()
    from xdsl.utils import marimo as xmo

    return xmo



from __future__ import annotations

from unittest.mock import Mock

import pytest

from xdsl.analysis.dataflow import (
    ChangeResult,
    DataFlowAnalysis,
    DataFlowSolver,
    LatticeAnchor,
    ProgramPoint,
)
from xdsl.analysis.dead_code_analysis import (
    CFGEdge,
    DeadCodeAnalysis,
    Executable,
    PredecessorState,
)
from xdsl.context import Context
from xdsl.dialects import arith, test
from xdsl.dialects.builtin import IntegerType, ModuleOp
from xdsl.ir import Block, Operation, Region
from xdsl.utils.test_value import create_ssa_value


def test_executable_initial_state():
    """Test the initial state of the Executable lattice element."""
    anchor = Mock(spec=LatticeAnchor)
    state = Executable(anchor)
    assert not state.live
    assert str(state) == "dead"
    assert not state.block_content_subscribers


def test_executable_set_to_live():
    """Test the set_to_live method."""
    anchor = Mock(spec=LatticeAnchor)
    state = Executable(anchor)

    # First time, should change
    assert state.set_to_live() is ChangeResult.CHANGE
    assert state.live
    assert str(state) == "live"

    # Second time, should not change
    assert state.set_to_live() is ChangeResult.NO_CHANGE
    assert state.live
    assert str(state) == "live"


@pytest.fixture
def op_in_block() -> tuple[Operation, Block]:
    op = test.TestOp()
    block = Block([op])
    return op, block


def test_executable_on_update_not_live(op_in_block: tuple[Operation, Block]):
    """Test that on_update does nothing if the state is not live."""
    op, _ = op_in_block
    solver = Mock(spec=DataFlowSolver)
    anchor = ProgramPoint.before(op)
    state = Executable(anchor)
    state.on_update(solver)
    solver.enqueue.assert_not_called()


def test_executable_on_update_not_program_point():
    """Test that on_update does nothing if the anchor is not a ProgramPoint."""
    solver = Mock(spec=DataFlowSolver)
    anchor = create_ssa_value(IntegerType(32))  # Not a ProgramPoint
    state = Executable(anchor)
    state.set_to_live()
    state.on_update(solver)
    solver.enqueue.assert_not_called()


def test_executable_on_update_not_start_of_block(op_in_block: tuple[Operation, Block]):
    """
    Test that on_update does nothing if the program point is not the start of a block.
    """
    op, block = op_in_block
    op2 = test.TestOp()
    block.add_op(op2)

    solver = Mock(spec=DataFlowSolver)
    # Program point after the first op, so not at the start of the block
    anchor = ProgramPoint.after(op)
    state = Executable(anchor)
    state.set_to_live()

    state.on_update(solver)
    solver.enqueue.assert_not_called()


def test_executable_on_update_start_of_block_no_subscribers(
    op_in_block: tuple[Operation, Block],
):
    """Test on_update at the start of a block with no subscribers."""
    op, _ = op_in_block
    solver = Mock(spec=DataFlowSolver)
    anchor = ProgramPoint.before(op)
    state = Executable(anchor)
    state.set_to_live()

    state.on_update(solver)
    solver.enqueue.assert_not_called()


def test_executable_on_update_start_of_block_with_subscribers():
    """Test on_update at the start of a block with subscribers."""
    # Setup a block with multiple ops
    op1 = test.TestOp()
    op2 = test.TestOp()
    _block = Block([op1, op2])

    solver = Mock(spec=DataFlowSolver)
    analysis1 = Mock(spec=DataFlowAnalysis)
    analysis2 = Mock(spec=DataFlowAnalysis)

    anchor = ProgramPoint.before(op1)
    state = Executable(anchor)
    state.set_to_live()
    state.block_content_subscribers.add(analysis1)
    state.block_content_subscribers.add(analysis2)

    state.on_update(solver)

    # Check that all ops are enqueued for all subscribed analyses
    assert solver.enqueue.call_count == 4
    solver.enqueue.assert_any_call((ProgramPoint.before(op1), analysis1))
    solver.enqueue.assert_any_call((ProgramPoint.before(op1), analysis2))
    solver.enqueue.assert_any_call((ProgramPoint.before(op2), analysis1))
    solver.enqueue.assert_any_call((ProgramPoint.before(op2), analysis2))


def test_executable_on_update_empty_block():
    """Test on_update for a live empty block."""
    block = Block()
    solver = Mock(spec=DataFlowSolver)
    analysis = Mock(spec=DataFlowAnalysis)

    # The anchor for an empty block is the block itself
    anchor = ProgramPoint.at_start_of_block(block)
    state = Executable(anchor)
    state.set_to_live()
    state.block_content_subscribers.add(analysis)

    # on_update checks if the anchor's entity is an Operation. For an empty
    # block, it's the block itself, so this should do nothing.
    state.on_update(solver)
    solver.enqueue.assert_not_called()


def test_cfg_edge_str():
    """Test the __str__ method of CFGEdge."""
    b1 = Block()
    b1.name_hint = "b1"
    b2 = Block()
    b2.name_hint = "b2"
    edge = CFGEdge(b1, b2)
    assert str(edge) == "edge(b1 -> b2)"


def test_cfg_edge_str_no_hint():
    """Test the __str__ method of CFGEdge when blocks have no name hints."""
    b1 = Block()
    b2 = Block()
    edge = CFGEdge(b1, b2)
    assert str(edge) == f"edge({id(b1)} -> {id(b2)})"


def test_predecessor_state_initial():
    """Test the initial state of PredecessorState."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)

    assert state.all_known is True
    assert state.known_predecessors == set()
    assert state.successor_inputs == {}
    assert str(state) == "Predecessors: [] (all known)"


def test_predecessor_state_set_has_unknown_predecessors():
    """Test the set_has_unknown_predecessors method."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)

    # First call should change the state
    assert state.set_has_unknown_predecessors() is ChangeResult.CHANGE
    assert state.all_known is False
    assert "unknown predecessors" in str(state)

    # Second call should not change the state
    assert state.set_has_unknown_predecessors() is ChangeResult.NO_CHANGE
    assert state.all_known is False


def test_predecessor_state_join_new_predecessor_no_inputs():
    """Test joining a new predecessor without inputs."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)
    pred_op = test.TestOp()

    assert state.join(pred_op) is ChangeResult.CHANGE
    assert state.known_predecessors == {pred_op}


def test_predecessor_state_join_existing_predecessor_no_inputs():
    """Test joining an existing predecessor without inputs."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)
    pred_op = test.TestOp()
    state.join(pred_op)

    assert state.join(pred_op) is ChangeResult.NO_CHANGE
    assert state.known_predecessors == {pred_op}


def test_predecessor_state_join_new_predecessor_with_inputs():
    """Test joining a new predecessor with inputs."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)
    pred_op = test.TestOp()
    inputs = (create_ssa_value(IntegerType(32)),)

    assert state.join(pred_op, inputs) is ChangeResult.CHANGE
    assert state.known_predecessors == {pred_op}
    assert state.successor_inputs == {pred_op: inputs}


def test_predecessor_state_join_existing_predecessor_same_inputs():
    """Test joining an existing predecessor with the same inputs."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)
    pred_op = test.TestOp()
    inputs = (create_ssa_value(IntegerType(32)),)
    state.join(pred_op, inputs)

    assert state.join(pred_op, inputs) is ChangeResult.NO_CHANGE
    assert state.known_predecessors == {pred_op}
    assert state.successor_inputs == {pred_op: inputs}


def test_predecessor_state_join_existing_predecessor_different_inputs():
    """Test joining an existing predecessor with different inputs."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)
    pred_op = test.TestOp()
    inputs1 = (create_ssa_value(IntegerType(32)),)
    state.join(pred_op, inputs1)

    inputs2 = (create_ssa_value(IntegerType(64)),)
    assert state.join(pred_op, inputs2) is ChangeResult.CHANGE
    assert state.known_predecessors == {pred_op}
    assert state.successor_inputs == {pred_op: inputs2}


def test_predecessor_state_join_existing_predecessor_add_inputs():
    """Test adding inputs to an existing predecessor that had none."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)
    pred_op = test.TestOp()
    state.join(pred_op)

    inputs = (create_ssa_value(IntegerType(32)),)
    assert state.join(pred_op, inputs) is ChangeResult.CHANGE
    assert state.known_predecessors == {pred_op}
    assert state.successor_inputs == {pred_op: inputs}


def test_predecessor_state_str():
    """Test the __str__ method of PredecessorState."""
    anchor = Mock(spec=LatticeAnchor)
    state = PredecessorState(anchor)

    pred1 = test.TestOp()
    state.join(pred1)
    assert str(state) == "Predecessors: [test.op] (all known)"

    pred2 = arith.ConstantOp.from_int_and_width(42, IntegerType(32))
    state.join(pred2)

    str_repr = str(state)
    assert "Predecessors: [" in str_repr
    # Order is not guaranteed
    assert "test.op" in str_repr
    assert "arith.constant" in str_repr
    assert "] (all known)" in str_repr

    state.set_has_unknown_predecessors()
    str_repr = str(state)
    assert "Predecessors: [" in str_repr
    assert "test.op" in str_repr
    assert "arith.constant" in str_repr
    assert "] (unknown predecessors)" in str_repr


@pytest.fixture
def solver_with_dca() -> tuple[DataFlowSolver, DeadCodeAnalysis]:
    """Fixture to get a solver with DeadCodeAnalysis loaded."""
    context = Context()
    solver = DataFlowSolver(context)
    analysis = solver.load(DeadCodeAnalysis)
    return solver, analysis


def test_dca_initialize_marks_entry_block_live(
    solver_with_dca: tuple[DataFlowSolver, DeadCodeAnalysis],
):
    """Test that DeadCodeAnalysis.initialize marks the entry block as live."""
    solver, analysis = solver_with_dca

    block = Block()
    region = Region([block])
    op = ModuleOp.create(regions=[region])

    solver.propagate_if_changed = Mock()

    analysis.initialize(op)

    entry_point = ProgramPoint.at_start_of_block(block)
    exec_state = solver.lookup_state(entry_point, Executable)
    assert exec_state is not None

    solver.propagate_if_changed.assert_called_once()
    call_args = solver.propagate_if_changed.call_args[0]
    assert call_args[0] is exec_state
    assert call_args[1] is ChangeResult.CHANGE
    assert exec_state.live


def test_dca_initialize_no_region(
    solver_with_dca: tuple[DataFlowSolver, DeadCodeAnalysis],
):
    """Test initialize on an op with no regions."""
    solver, analysis = solver_with_dca
    op = test.TestOp()  # has no regions by default

    solver.propagate_if_changed = Mock()
    analysis.initialize(op)
    solver.propagate_if_changed.assert_not_called()


def test_dca_initialize_empty_region(
    solver_with_dca: tuple[DataFlowSolver, DeadCodeAnalysis],
):
    """Test initialize on an op with an empty region."""
    solver, analysis = solver_with_dca
    region = Region()
    op = ModuleOp.create(regions=[region])

    solver.propagate_if_changed = Mock()
    analysis.initialize(op)
    solver.propagate_if_changed.assert_not_called()


def test_dca_visit_op_in_live_block(
    solver_with_dca: tuple[DataFlowSolver, DeadCodeAnalysis],
):
    """Test visiting an op in a live block."""
    solver, analysis = solver_with_dca

    op = test.TestOp()
    block = Block([op])

    block_point = ProgramPoint.at_start_of_block(block)
    exec_state = solver.get_or_create_state(block_point, Executable)
    exec_state.set_to_live()

    op_point = ProgramPoint.before(op)
    analysis.visit(op_point)

    assert (op_point, analysis) in exec_state.dependents
    assert analysis in exec_state.block_content_subscribers


def test_dca_visit_op_in_dead_block(
    solver_with_dca: tuple[DataFlowSolver, DeadCodeAnalysis],
):
    """Test visiting an op in a dead block."""
    solver, analysis = solver_with_dca

    op = test.TestOp()
    block = Block([op])

    block_point = ProgramPoint.at_start_of_block(block)
    exec_state = solver.get_or_create_state(block_point, Executable)

    op_point = ProgramPoint.before(op)
    analysis.visit(op_point)

    assert (op_point, analysis) in exec_state.dependents
    assert analysis not in exec_state.block_content_subscribers


def test_dca_visit_non_op_point(
    solver_with_dca: tuple[DataFlowSolver, DeadCodeAnalysis],
):
    """Test visiting a non-op program point."""
    solver, analysis = solver_with_dca

    block = Block()
    point = ProgramPoint.at_end_of_block(block)  # Not an op

    solver.get_or_create_state = Mock()
    analysis.visit(point)
    solver.get_or_create_state.assert_not_called()



# RUN: python %s | filecheck %s

from xdsl.dialects import bigint, builtin
from xdsl.frontend.pyast.context import PyASTContext
from xdsl.frontend.pyast.utils.exceptions import FrontendProgramException

ctx = PyASTContext()
ctx.register_type(int, bigint.bigint)
ctx.register_type(bool, builtin.i1)


@ctx.parse_program
def redefined_function():  # pyright: ignore[reportRedeclaration]
    return


@ctx.parse_program
def redefined_function(x: int) -> int:
    return x


# Function definitions can be overriden
print(redefined_function.module)
# CHECK:      builtin.module {
# CHECK-NEXT:     func.func @redefined_function(%x : !bigint.bigint) -> !bigint.bigint {
# CHECK-NEXT:       func.return %x : !bigint.bigint
# CHECK-NEXT:     }
# CHECK-NEXT:   }


# CHECK-NEXT: Cannot have an inner function 'inner' inside another function.
@ctx.parse_program
def outer():
    def inner():  # pyright: ignore[reportUnusedFunction]
        return

    return


try:
    outer.module
except FrontendProgramException as e:
    print(e.msg)


# CHECK-NEXT: Expected non-zero number of return types in function 'missing_return_value', but got 0.
@ctx.parse_program
def missing_return_value() -> int:
    return  # pyright: ignore[reportReturnType]


try:
    missing_return_value.module
except FrontendProgramException as e:
    print(e.msg)


try:
    # CHECK-NEXT: Cannot re-register type name 'int'
    ctx.register_type(int, bigint.bigint)
except FrontendProgramException as e:
    print(e.msg)


try:
    # CHECK-NEXT: Cannot re-register function 'int.__add__'
    ctx.register_function(int.__add__, bigint.AddOp)
    ctx.register_function(int.__add__, bigint.AddOp)
except FrontendProgramException as e:
    print(e.msg)


try:
    # CHECK-NEXT: Cannot register multiple source types for IR type '!bigint.bigint'
    ctx.register_type(float, bigint.bigint)
except FrontendProgramException as e:
    print(e.msg)


try:
    # CHECK-NEXT: Cannot register multiple source types for IR type '!bigint.bigint'
    ctx.register_type(float, bigint.bigint)
except FrontendProgramException as e:
    print(e.msg)


# CHECK: Expected non-zero number of return types in function 'test_no_return_type', but got 0.
@ctx.parse_program
def test_no_return_type(a: int) -> int:
    return  # pyright: ignore[reportReturnType]


try:
    test_no_return_type.module
except FrontendProgramException as e:
    print(e.msg)


# CHECK: Type signature and the type of the return value do not match at position 0: expected i1, got !bigint.bigint.
@ctx.parse_program
def test_wrong_return_type(a: bool, b: int) -> bool:
    return b  # pyright: ignore[reportReturnType]


try:
    test_wrong_return_type.module
except FrontendProgramException as e:
    print(e.msg)


# CHECK: Expected no return types in function 'test_no_return_types'.
@ctx.parse_program
def test_no_return_types(a: int):
    return a


try:
    test_no_return_types.module
except FrontendProgramException as e:
    print(e.msg)


# CHECK: Expected the same types for binary operation 'Add', but got !bigint.bigint and i1.
@ctx.parse_program
def bin_op_type_mismatch(a: int, b: bool) -> int:
    return a + b


try:
    bin_op_type_mismatch.module
except FrontendProgramException as e:
    print(e.msg)


# CHECK: Expected the same types for comparison operator 'Lt', but got !bigint.bigint and i1.
@ctx.parse_program
def cmp_op_type_mismatch(a: int, b: bool) -> bool:
    return a < b


try:
    cmp_op_type_mismatch.module
except FrontendProgramException as e:
    print(e.msg)



from io import StringIO

import marimo as mo

from xdsl.dialects import builtin
from xdsl.interpreter import Interpreter
from xdsl.interpreters.arith import ArithFunctions
from xdsl.interpreters.printf import PrintfFunctions
from xdsl.interpreters.scf import ScfFunctions
from xdsl.interpreters.tensor import TensorFunctions


def interp(module: builtin.ModuleOp) -> str:
    _io = StringIO()

    _i = Interpreter(module=module, file=_io)
    _i.register_implementations(ArithFunctions())
    _i.register_implementations(ScfFunctions())
    _i.register_implementations(PrintfFunctions())
    _i.register_implementations(TensorFunctions())
    _i.run_ssacfg_region(module.body, ())

    # Lowercase to avoid capitals in `True` and `False` printing of bools.
    # Safe since we never print strings other than `True` and `False`.
    return _io.getvalue().lower()


def rust_md(code: str) -> mo.Html:
    return mo.md("`" * 3 + "rust\n" + code + "\n" + "`" * 3)



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from typing_extensions import Self

from xdsl.analysis.dataflow import (
    ChangeResult,
    DataFlowAnalysis,
    DataFlowSolver,
    ProgramPoint,
)
from xdsl.analysis.dead_code_analysis import Executable
from xdsl.analysis.sparse_analysis import (
    AbstractLatticeValue,
    Lattice,
    PropagatingLattice,
    SparseForwardDataFlowAnalysis,
)
from xdsl.context import Context
from xdsl.dialects import test
from xdsl.dialects.builtin import IntegerType, ModuleOp, i32
from xdsl.ir import Block, Operation, Region, SSAValue
from xdsl.utils.test_value import create_ssa_value

# region Test lattice implementations


@dataclass(frozen=True)
class SimpleLatticeValue:
    """A simple lattice value for testing AbstractLatticeValue protocol."""

    value: int

    def meet(self, other: SimpleLatticeValue) -> SimpleLatticeValue:
        return SimpleLatticeValue(min(self.value, other.value))

    def join(self, other: SimpleLatticeValue) -> SimpleLatticeValue:
        return SimpleLatticeValue(max(self.value, other.value))


class SimpleSparseLattice(PropagatingLattice):
    """Concrete implementation of SparseLatticeSubscriberBase for testing."""

    def __init__(self, anchor: SSAValue):
        super().__init__(anchor)
        self.value = 0

    def join(self, other: Self) -> ChangeResult:
        if other.value > self.value:
            self.value = other.value
            return ChangeResult.CHANGE
        return ChangeResult.NO_CHANGE

    def meet(self, other: Self) -> ChangeResult:
        if other.value < self.value:
            self.value = other.value
            return ChangeResult.CHANGE
        return ChangeResult.NO_CHANGE

    def __str__(self) -> str:
        return f"SimpleSparseLattice({self.value})"


# endregion


# region AbstractLatticeValue protocol tests


def test_abstract_lattice_value_meet():
    """Test that meet operation works on lattice values."""
    val1 = SimpleLatticeValue(5)
    val2 = SimpleLatticeValue(3)
    result = val1.meet(val2)
    assert result.value == 3


def test_abstract_lattice_value_join():
    """Test that join operation works on lattice values."""
    val1 = SimpleLatticeValue(5)
    val2 = SimpleLatticeValue(3)
    result = val1.join(val2)
    assert result.value == 5


# endregion


# region SparseLatticeSubscriberBase tests


def test_use_def_subscribe():
    """Test that analyses can subscribe to use-def updates."""
    anchor = create_ssa_value(IntegerType(32))
    lattice = SimpleSparseLattice(anchor)

    assert lattice.anchor is anchor
    assert len(lattice.use_def_subscribers) == 0
    assert len(lattice.dependents) == 0

    solver = DataFlowSolver(Context())

    class DummyAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            pass

        def visit(self, point: ProgramPoint) -> None:
            pass

    analysis1 = DummyAnalysis(solver)
    analysis2 = DummyAnalysis(solver)

    lattice.use_def_subscribe(analysis1)
    assert analysis1 in lattice.use_def_subscribers
    assert len(lattice.use_def_subscribers) == 1

    lattice.use_def_subscribe(analysis2)
    assert analysis2 in lattice.use_def_subscribers
    assert len(lattice.use_def_subscribers) == 2

    # Subscribing the same analysis twice shouldn't duplicate
    lattice.use_def_subscribe(analysis1)
    assert len(lattice.use_def_subscribers) == 2


def test_on_update_propagates_to_users():
    """Test that on_update enqueues work for users of the SSA value."""
    # Create operations with use-def chain
    producer = test.TestOp(result_types=[i32])
    consumer1 = test.TestOp(operands=[producer.results[0]])
    consumer2 = test.TestOp(operands=[producer.results[0]])

    Block([producer, consumer1, consumer2])

    # Set up analysis and solver
    solver = DataFlowSolver(Context())
    visited_points: list[ProgramPoint] = []

    class TrackingAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            # Create lattice and subscribe analysis
            lattice = SimpleSparseLattice(producer.results[0])
            lattice.use_def_subscribe(self)
            # Simulate a change to trigger the update
            lattice.on_update(solver)

        def visit(self, point: ProgramPoint) -> None:
            visited_points.append(point)

    solver.load(TrackingAnalysis)

    # Run solver
    solver.initialize_and_run(producer)

    # Check that both users were visited
    assert len(visited_points) == 2
    assert any(p.op is consumer1 for p in visited_points)
    assert any(p.op is consumer2 for p in visited_points)


def test_on_update_calls_super():
    """Test that on_update also propagates to explicit dependents."""
    anchor = create_ssa_value(IntegerType(32))
    lattice = SimpleSparseLattice(anchor)

    solver = DataFlowSolver(Context())
    visited_points: list[ProgramPoint] = []

    class TrackingAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            # Add explicit dependent
            dependent_op = test.TestOp()
            dependent_point = ProgramPoint.before(dependent_op)
            lattice.dependents.add((dependent_point, self))
            # Simulate a change to trigger the update
            lattice.on_update(solver)

        def visit(self, point: ProgramPoint) -> None:
            visited_points.append(point)

    solver.load(TrackingAnalysis)

    # Create a dummy operation to initialize with
    dummy_op = test.TestOp()

    # Run solver
    solver.initialize_and_run(dummy_op)

    # Check that explicit dependent was visited
    assert len(visited_points) == 1


def test_on_update_with_non_ssa_anchor():
    """Test that on_update handles non-SSAValue anchors gracefully."""
    # Create a dummy SSAValue for the test since SimpleSparseLattice requires SSAValue
    anchor = create_ssa_value(IntegerType(32))
    lattice = SimpleSparseLattice(anchor)

    solver = DataFlowSolver(Context())

    class DummyAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            pass

        def visit(self, point: ProgramPoint) -> None:
            pass

    analysis = DummyAnalysis(solver)
    lattice.use_def_subscribe(analysis)

    # Create a dummy operation to initialize with
    dummy_op = test.TestOp()

    # Should not raise an error
    solver.initialize_and_run(dummy_op)


def test_sparse_lattice_join_and_meet():
    """Test that join and meet operations work and return correct ChangeResult."""
    anchor = create_ssa_value(IntegerType(32))
    lattice1 = SimpleSparseLattice(anchor)
    lattice2 = SimpleSparseLattice(anchor)

    lattice1.value = 5
    lattice2.value = 3

    # Join should take maximum and report change
    result = lattice1.join(lattice2)
    assert result == ChangeResult.NO_CHANGE  # 5 >= 3
    assert lattice1.value == 5

    lattice2.value = 10
    result = lattice1.join(lattice2)
    assert result == ChangeResult.CHANGE  # 5 < 10
    assert lattice1.value == 10

    # Meet should take minimum and report change
    lattice1.value = 5
    lattice2.value = 3
    result = lattice1.meet(lattice2)
    assert result == ChangeResult.CHANGE  # 5 > 3
    assert lattice1.value == 3

    result = lattice1.meet(lattice2)
    assert result == ChangeResult.NO_CHANGE  # 3 == 3
    assert lattice1.value == 3


# endregion


# region Test lattice value implementations


@dataclass(frozen=True)
class TestLatticeValue(AbstractLatticeValue):
    """A test lattice value implementing AbstractLatticeValue protocol."""

    value: int

    @classmethod
    def initial_value(cls) -> Self:
        return cls(-1)

    def meet(self, other: TestLatticeValue) -> TestLatticeValue:
        """Meet returns minimum value."""
        return TestLatticeValue(min(self.value, other.value))

    def join(self, other: TestLatticeValue) -> TestLatticeValue:
        """Join returns maximum value."""
        return TestLatticeValue(max(self.value, other.value))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TestLatticeValue):
            return NotImplemented
        return self.value == other.value


class TestLattice(Lattice[TestLatticeValue]):
    """Concrete Lattice implementation for testing."""

    value_cls = TestLatticeValue


# endregion


# region Lattice initialization tests


def test_lattice_initialization_with_value():
    """Test that Lattice initializes correctly with an explicit value."""
    anchor = create_ssa_value(IntegerType(32))
    initial_value = TestLatticeValue(42)
    lattice = TestLattice(anchor, value=initial_value)

    assert lattice.anchor is anchor
    assert lattice.value == initial_value
    assert lattice.value.value == 42
    assert len(lattice.use_def_subscribers) == 0


def test_lattice_initialization_without_value():
    """Test that Lattice initializes with default value when none provided."""
    anchor = create_ssa_value(IntegerType(32))
    lattice = TestLattice(anchor)

    assert lattice.anchor is anchor
    # Should initialize with default constructor of TestLatticeValue
    assert isinstance(lattice.value, TestLatticeValue)
    assert not lattice.use_def_subscribers


# endregion


# region Lattice value property tests


def test_lattice_value_property():
    """Test that the value property correctly exposes the internal value."""
    anchor = create_ssa_value(IntegerType(32))
    initial_value = TestLatticeValue(100)
    lattice = TestLattice(anchor, value=initial_value)

    # Value property should return the internal value
    assert lattice.value is lattice._value  # pyright: ignore[reportPrivateUsage]
    assert lattice.value == initial_value


# endregion


# region Lattice meet operation tests


def test_lattice_meet_with_change():
    """Test that meet operation detects changes and updates value."""
    anchor = create_ssa_value(IntegerType(32))
    lattice1 = TestLattice(anchor, value=TestLatticeValue(5))
    lattice2 = TestLattice(anchor, value=TestLatticeValue(3))

    # Meet should take minimum: min(5, 3) = 3
    result = lattice1.meet(lattice2)

    assert result == ChangeResult.CHANGE
    assert lattice1.value.value == 3


def test_lattice_meet_without_change():
    """Test that meet operation detects when no change occurs."""
    anchor = create_ssa_value(IntegerType(32))
    lattice1 = TestLattice(anchor, value=TestLatticeValue(3))
    lattice2 = TestLattice(anchor, value=TestLatticeValue(5))

    # Meet should take minimum: min(3, 5) = 3 (no change)
    result = lattice1.meet(lattice2)

    assert result == ChangeResult.NO_CHANGE
    assert lattice1.value.value == 3


# endregion


# region Lattice join operation tests


def test_lattice_join_with_change():
    """Test that join operation detects changes and updates value."""
    anchor = create_ssa_value(IntegerType(32))
    lattice1 = TestLattice(anchor, value=TestLatticeValue(5))
    lattice2 = TestLattice(anchor, value=TestLatticeValue(10))

    # Join should take maximum: max(5, 10) = 10
    result = lattice1.join(lattice2)

    assert result == ChangeResult.CHANGE
    assert lattice1.value.value == 10


def test_lattice_join_without_change():
    """Test that join operation detects when no change occurs."""
    anchor = create_ssa_value(IntegerType(32))
    lattice1 = TestLattice(anchor, value=TestLatticeValue(10))
    lattice2 = TestLattice(anchor, value=TestLatticeValue(5))

    # Join should take maximum: max(10, 5) = 10 (no change)
    result = lattice1.join(lattice2)

    assert result == ChangeResult.NO_CHANGE
    assert lattice1.value.value == 10


# endregion


# region Lattice propagation tests


def test_lattice_use_def_subscription():
    """Test that analyses can subscribe to Lattice updates."""
    anchor = create_ssa_value(IntegerType(32))
    lattice = TestLattice(anchor, value=TestLatticeValue(0))

    solver = DataFlowSolver(Context())

    class DummyAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            pass

        def visit(self, point: ProgramPoint) -> None:
            pass

    analysis1 = DummyAnalysis(solver)
    analysis2 = DummyAnalysis(solver)

    lattice.use_def_subscribe(analysis1)
    assert analysis1 in lattice.use_def_subscribers

    lattice.use_def_subscribe(analysis2)
    assert len(lattice.use_def_subscribers) == 2


def test_lattice_propagates_on_join_change():
    """Test that join changes trigger propagation to users."""
    # Create operations with use-def chain
    producer = test.TestOp(result_types=[i32])
    consumer1 = test.TestOp(operands=[producer.results[0]])
    consumer2 = test.TestOp(operands=[producer.results[0]])

    Block([producer, consumer1, consumer2])

    solver = DataFlowSolver(Context())
    visited_points: list[ProgramPoint] = []

    class PropagationAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            if op is producer:
                lattice = TestLattice(producer.results[0], value=TestLatticeValue(0))
                lattice.use_def_subscribe(self)
                # Store for later access
                self.lattice = lattice
                # Enqueue initial work at the producer
                solver.enqueue((ProgramPoint.before(producer), self))

        def visit(self, point: ProgramPoint) -> None:
            visited_points.append(point)
            # Simulate a change via join if we're at the producer
            if point.op is producer and hasattr(self, "lattice"):
                other = TestLattice(producer.results[0], value=TestLatticeValue(10))
                result = self.lattice.join(other)
                # Propagate the change to trigger visiting consumers
                self.propagate_if_changed(self.lattice, result)

    solver.load(PropagationAnalysis)
    solver.initialize_and_run(producer)

    # Should visit producer, then consumers after the change
    assert len(visited_points) == 3  # producer + 2 consumers
    assert any(p.op is consumer1 for p in visited_points)
    assert any(p.op is consumer2 for p in visited_points)


def test_lattice_propagates_on_meet_change():
    """Test that meet changes trigger propagation to users."""
    # Create operations with use-def chain
    producer = test.TestOp(result_types=[i32])
    consumer = test.TestOp(operands=[producer.results[0]])

    Block([producer, consumer])

    solver = DataFlowSolver(Context())
    visited_points: list[ProgramPoint] = []

    class PropagationAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            if op is producer:
                lattice = TestLattice(producer.results[0], value=TestLatticeValue(10))
                lattice.use_def_subscribe(self)
                self.lattice = lattice
                # Enqueue initial work at the producer
                solver.enqueue((ProgramPoint.before(producer), self))

        def visit(self, point: ProgramPoint) -> None:
            visited_points.append(point)
            # Simulate a change via meet if we're at the producer
            if point.op is producer and hasattr(self, "lattice"):
                other = TestLattice(producer.results[0], value=TestLatticeValue(5))
                result = self.lattice.meet(other)
                # Propagate the change to trigger visiting consumers
                self.propagate_if_changed(self.lattice, result)

    solver.load(PropagationAnalysis)
    solver.initialize_and_run(producer)

    # Should visit producer, then consumer after the change
    assert len(visited_points) == 2  # producer + consumer
    assert any(p.op is consumer for p in visited_points)


# endregion


# region Lattice string representation tests


def test_lattice_string_representation():
    """Test that Lattice has a meaningful string representation."""
    anchor = create_ssa_value(IntegerType(32))
    lattice = TestLattice(anchor, value=TestLatticeValue(42))

    str_repr = str(lattice)
    assert "42" in str_repr


# endregion


# region Lattice integration tests


def test_lattice_multiple_operations():
    """Test Lattice through multiple operations."""
    anchor = create_ssa_value(IntegerType(32))
    lattice = TestLattice(anchor, value=TestLatticeValue(5))

    other1 = TestLattice(anchor, value=TestLatticeValue(10))
    other2 = TestLattice(anchor, value=TestLatticeValue(3))

    # First join: 5 ∨ 10 = 10
    result1 = lattice.join(other1)
    assert result1 == ChangeResult.CHANGE
    assert lattice.value.value == 10

    # Second join: 10 ∨ 3 = 10 (no change)
    result2 = lattice.join(other2)
    assert result2 == ChangeResult.NO_CHANGE
    assert lattice.value.value == 10

    # Meet: 10 ∧ 3 = 3
    result3 = lattice.meet(other2)
    assert result3 == ChangeResult.CHANGE
    assert lattice.value.value == 3


# endregion


# region Test lattice for simple analysis


@dataclass(frozen=True)
class KnownLatticeValue(AbstractLatticeValue):
    """A simple lattice value: Unknown (⊥) < Known < Top (⊤)"""

    value: Literal["unknown", "top"] | int

    @classmethod
    def initial_value(cls) -> Self:
        return cls("unknown")

    def meet(self, other: KnownLatticeValue) -> KnownLatticeValue:
        """Meet returns the more precise value."""
        match (self.value, other.value):
            case ("unknown", _):
                return other
            case (_, "unknown"):
                return self
            case ("top", _) | (_, "top"):
                return KnownLatticeValue("top")
            case (a, b) if a == b:
                return self
            case _:
                return KnownLatticeValue("top")

    def join(self, other: KnownLatticeValue) -> KnownLatticeValue:
        """Join returns the less precise value."""
        match (self.value, other.value):
            case ("top", _) | (_, "top"):
                return KnownLatticeValue("top")
            case ("unknown", _):
                return other
            case (_, "unknown"):
                return self
            case (a, b) if a == b:
                return self
            case _:
                return KnownLatticeValue("top")


class SimpleLattice(Lattice[KnownLatticeValue]):
    value_cls = KnownLatticeValue


# endregion


# region Concrete analysis implementation


class SimpleForwardAnalysis(SparseForwardDataFlowAnalysis[SimpleLattice]):
    """A simple forward analysis for testing."""

    def __init__(self, solver: DataFlowSolver):
        super().__init__(solver, SimpleLattice)
        self.visited_ops: list[Operation] = []

    def visit_operation_impl(
        self,
        op: Operation,
        operands: list[SimpleLattice],
        results: list[SimpleLattice],
    ) -> None:
        """Simple transfer function: propagate operand states to results."""
        self.visited_ops.append(op)

        # For testing, just copy first operand state to all results
        if operands and results:
            for result in results:
                self.join(result, operands[0])

    def set_to_entry_state(self, lattice: SimpleLattice) -> None:
        """Set lattice to top (most pessimistic) state."""
        lattice._value = KnownLatticeValue("top")  # pyright: ignore[reportPrivateUsage]


# endregion


# region Basic initialization tests


def test_forward_analysis_initialization():
    """Test that initialization creates lattices for region arguments."""
    ctx = Context()
    solver = DataFlowSolver(ctx)
    analysis = SimpleForwardAnalysis(solver)

    # Create operation with a region containing arguments
    arg_type = IntegerType(32)
    block = Block(arg_types=[arg_type, arg_type])
    region = Region([block])
    op = ModuleOp(region)

    # Mark the entry block as executable
    entry_point = ProgramPoint.at_start_of_block(block)
    executable = solver.get_or_create_state(entry_point, Executable)
    executable.live = True

    analysis.initialize(op)

    # Check that lattices were created for arguments
    arg0_lattice = solver.lookup_state(block.args[0], SimpleLattice)
    arg1_lattice = solver.lookup_state(block.args[1], SimpleLattice)

    assert arg0_lattice is not None
    assert arg1_lattice is not None
    # Entry state should be "top"
    assert arg0_lattice.value.value == "top"
    assert arg1_lattice.value.value == "top"


# endregion


# region visit_operation tests


def test_visit_operation_creates_lattices():
    """Test that visit_operation retrieves/creates lattices for operands and results."""
    ctx = Context()
    solver = DataFlowSolver(ctx)
    analysis = SimpleForwardAnalysis(solver)

    # Create operations
    producer = test.TestOp(result_types=[i32])
    consumer = test.TestOp(operands=[producer.results[0]], result_types=[i32])
    block = Block([producer, consumer])

    # Mark block as executable
    entry_point = ProgramPoint.at_start_of_block(block)
    executable = solver.get_or_create_state(entry_point, Executable)
    executable.live = True

    solver._is_running = True  # pyright: ignore[reportPrivateUsage]

    # Visit consumer operation
    analysis.visit_operation(consumer)

    # Check that lattices exist
    operand_lattice = solver.lookup_state(producer.results[0], SimpleLattice)
    result_lattice = solver.lookup_state(consumer.results[0], SimpleLattice)

    assert operand_lattice is not None
    assert result_lattice is not None


def test_visit_operation_skips_non_executable_blocks():
    """Test that operations in non-executable blocks are skipped."""
    ctx = Context()
    solver = DataFlowSolver(ctx)
    analysis = SimpleForwardAnalysis(solver)

    # Create operation in a block
    op = test.TestOp(result_types=[i32])
    block = Block([op])

    # Don't mark block as executable
    entry_point = ProgramPoint.at_start_of_block(block)
    executable = solver.get_or_create_state(entry_point, Executable)
    executable.live = False

    # Visit should not process the operation
    analysis.visit_operation(op)

    # The operation should not be in visited list
    assert op not in analysis.visited_ops


def test_visit_operation_subscribes_to_operands():
    """Test that visit_operation subscribes to operand changes."""
    ctx = Context()
    solver = DataFlowSolver(ctx)
    analysis = SimpleForwardAnalysis(solver)

    # Create operations
    producer = test.TestOp(result_types=[i32])
    consumer = test.TestOp(operands=[producer.results[0]], result_types=[i32])
    block = Block([producer, consumer])

    # Mark block as executable
    entry_point = ProgramPoint.at_start_of_block(block)
    executable = solver.get_or_create_state(entry_point, Executable)
    executable.live = True

    solver._is_running = True  # pyright: ignore[reportPrivateUsage]

    # Visit consumer
    analysis.visit_operation(consumer)

    # Check that analysis subscribed to the operand lattice
    operand_lattice = analysis.get_lattice_element(producer.results[0])
    assert analysis in operand_lattice.use_def_subscribers


def test_visit_operation_handles_no_results():
    """Test that operations without results are handled correctly."""
    ctx = Context()
    solver = DataFlowSolver(ctx)
    analysis = SimpleForwardAnalysis(solver)

    # Create operation without results
    op = test.TestOp(result_types=[])
    block = Block([op])

    # Mark block as executable
    entry_point = ProgramPoint.at_start_of_block(block)
    executable = solver.get_or_create_state(entry_point, Executable)
    executable.live = True

    # Should return early without error
    analysis.visit_operation(op)
    assert op not in analysis.visited_ops


# endregion


# region join operation tests


def test_join_operation():
    """Test that join operation works correctly."""
    ctx = Context()
    solver = DataFlowSolver(ctx)
    analysis = SimpleForwardAnalysis(solver)

    # Create two lattices with the same anchor
    from xdsl.utils.test_value import create_ssa_value

    anchor = create_ssa_value(IntegerType(32))
    lattice1 = SimpleLattice(anchor, value=KnownLatticeValue("unknown"))
    lattice2 = SimpleLattice(anchor, value=KnownLatticeValue(42))

    solver._is_running = True  # pyright: ignore[reportPrivateUsage]

    # Join should update lattice1
    analysis.join(lattice1, lattice2)

    assert lattice1.value.value == 42


# endregion


# region Integration test


def test_simple_forward_analysis_integration():
    """Test a simple forward analysis end-to-end."""
    ctx = Context()
    solver = DataFlowSolver(ctx)
    analysis = solver.load(SimpleForwardAnalysis)

    # Create a simple chain: producer -> consumer
    producer = test.TestOp(result_types=[i32])
    consumer = test.TestOp(operands=[producer.results[0]], result_types=[i32])
    block = Block([producer, consumer])
    region = Region([block])
    module = ModuleOp(region)

    # Mark block as executable before running
    entry_point = ProgramPoint.at_start_of_block(block)
    executable = solver.get_or_create_state(entry_point, Executable)
    executable.live = True

    # Run the analysis
    solver.initialize_and_run(module)

    # Both operations should have been visited
    assert producer in analysis.visited_ops
    assert consumer in analysis.visited_ops

    # Lattices should exist for results
    producer_lattice = solver.lookup_state(producer.results[0], SimpleLattice)
    consumer_lattice = solver.lookup_state(consumer.results[0], SimpleLattice)

    assert producer_lattice is not None
    assert consumer_lattice is not None


# endregion



# RUN: python %s | filecheck %s


from xdsl.dialects import bigint, builtin
from xdsl.frontend.pyast.context import PyASTContext
from xdsl.frontend.pyast.utils.exceptions import CodeGenerationException

ctx = PyASTContext(post_transforms=[])
ctx.register_type(bool, builtin.i1)
ctx.register_type(int, bigint.bigint)


# CHECK: cf.assert %{{.*}}, ""
@ctx.parse_program
def test_assert_I(cond: bool):
    assert cond
    return


print(test_assert_I.module)


# CHECK: cf.assert %{{.*}}, "some message"
@ctx.parse_program
def test_assert_II(cond: bool):
    assert cond, "some message"
    return


print(test_assert_II.module)


# CHECK: Expected a string constant for assertion message, found 'ast.Name'
@ctx.parse_program
def test_assert_message_type(cond: bool, a: int):
    assert cond, a
    return


try:
    test_assert_message_type.module
except CodeGenerationException as e:
    print(e.msg)



import pytest

from xdsl.utils.disjoint_set import DisjointSet, IntDisjointSet


def test_disjoint_set_init():
    ds = IntDisjointSet(size=5)
    assert ds.value_count() == 5
    # Each element should start in its own set
    for i in range(5):
        assert ds[i] == i


def test_disjoint_set_add():
    ds = IntDisjointSet(size=2)
    assert ds.value_count() == 2

    new_val = ds.add()
    assert new_val == 2
    assert ds.value_count() == 3
    assert ds[new_val] == new_val


def test_disjoint_set_find_invalid():
    ds = IntDisjointSet(size=3)
    with pytest.raises(KeyError):
        ds[3]
    with pytest.raises(KeyError):
        ds[-1]


def test_disjoint_set_union():
    ds = IntDisjointSet(size=4)

    # Union 0 and 1
    assert ds.union(0, 1)
    root = ds[0]
    assert ds[1] == root
    assert ds.connected(0, 1)
    assert not ds.connected(0, 2)

    # Union 2 and 3
    assert ds.union(2, 3)
    root2 = ds[2]
    assert ds[3] == root2
    assert ds.connected(2, 3)
    assert not ds.connected(1, 2)

    # Union already connected elements
    assert not ds.union(0, 1)
    assert ds.connected(0, 1)

    # Union two sets
    assert ds.union(1, 2)
    final_root = ds[0]
    assert ds[1] == final_root
    assert ds[2] == final_root
    assert ds[3] == final_root
    # After unioning all elements, they should all be connected
    assert ds.connected(0, 1)
    assert ds.connected(1, 2)
    assert ds.connected(2, 3)
    assert ds.connected(0, 3)


def test_disjoint_set_path_compression():
    ds = IntDisjointSet(size=4)

    # Create a chain: 3->2->1->0
    ds._parent = [0, 0, 1, 2]  # pyright: ignore[reportPrivateUsage]
    ds._count = [4, 3, 2, 1]  # pyright: ignore[reportPrivateUsage]

    # Find should compress the path
    root = ds[3]
    # After compression, all nodes should point directly to root
    assert ds._parent[3] == root  # pyright: ignore[reportPrivateUsage]
    assert ds._parent[2] == root  # pyright: ignore[reportPrivateUsage]
    assert ds._parent[1] == root  # pyright: ignore[reportPrivateUsage]
    assert ds._parent[0] == root  # pyright: ignore[reportPrivateUsage]


def test_generic_disjoint_set():
    ds = DisjointSet(["a", "b", "c", "d"])

    # Union a and b
    assert ds.union("a", "b")
    root = ds.find("a")
    assert ds.find("b") == root
    assert ds.connected("a", "b")
    assert not ds.connected("a", "c")

    # Union c and d
    assert ds.union("c", "d")
    root2 = ds.find("c")
    assert ds.find("d") == root2
    assert ds.connected("c", "d")
    assert not ds.connected("b", "c")

    # Union already connected elements
    assert not ds.union("a", "b")
    assert ds.connected("a", "b")

    # Union two sets
    assert ds.union("b", "c")
    final_root = ds.find("a")
    assert ds.find("b") == final_root
    assert ds.find("c") == final_root
    assert ds.find("d") == final_root
    # After unioning all elements, they should all be connected
    assert ds.connected("a", "b")
    assert ds.connected("b", "c")
    assert ds.connected("c", "d")
    assert ds.connected("a", "d")


def test_generic_disjoint_set_add():
    ds = DisjointSet(["a", "b"])
    ds.add("c")
    ds.add("d")

    assert ds.union("a", "c")
    root = ds.find("a")
    assert ds.find("c") == root

    assert ds.union("b", "d")
    root2 = ds.find("b")
    assert ds.find("d") == root2


def test_generic_disjoint_set_find_invalid():
    ds = DisjointSet(["a", "b", "c"])
    with pytest.raises(KeyError):
        ds.find("d")


def test_generic_disjoint_set_union_by_size():
    ds: DisjointSet[str] = DisjointSet(["a", "b", "c", "d"])
    ds.union("a", "b")
    ds.union("a", "c")
    # The set containing "a" is larger than the one containing "c"
    # so the canonical representative should be "a".
    assert ds.find("a") == "a"
    assert ds.find("c") == "a"

    ds.union("d", "c")
    assert ds.find("d") == "a"
    assert ds.find("c") == "a"


def test_int_disjoint_set_rooted_union():
    ds = IntDisjointSet(size=6)

    # Test successful rooted union
    assert ds.union_left(0, 1)  # Union with 0 as root
    assert ds[1] == 0
    assert ds.connected(0, 1)

    # Test rooted union with multiple elements
    assert ds.union_left(0, 2)  # Add 2 to the set rooted at 0
    assert ds[2] == 0
    assert ds.connected(0, 2)
    assert ds.connected(1, 2)

    # Test rooted union when sets are already connected
    assert not ds.union_left(0, 1)  # Should return False
    assert ds[1] == 0  # Still connected

    # Test error when representative is not a root
    ds.union_left(1, 3)  # 1 is not a root, 0 is its root
    assert ds[3] == 0
    assert ds.connected(1, 3)

    # Test that the specified representative becomes the root
    assert ds.union_left(4, 5)  # 4 becomes root of its own set with 5
    assert ds[4] == 4
    assert ds[5] == 4

    # Now union the two sets with 0 as the specified root
    assert ds.union_left(0, 4)
    assert ds[4] == 0  # 4's set is now under 0
    assert ds[5] == 0  # 5 also points to 0
    assert ds.connected(0, 5)


def test_generic_disjoint_set_rooted_union():
    ds = DisjointSet(["a", "b", "c", "d", "e", "f"])

    # Test successful rooted union
    assert ds.union_left("a", "b")  # Union with "a" as root
    assert ds.find("b") == "a"
    assert ds.connected("a", "b")

    # Test rooted union with multiple elements
    assert ds.union_left("a", "c")  # Add "c" to the set rooted at "a"
    assert ds.find("c") == "a"
    assert ds.connected("a", "c")
    assert ds.connected("b", "c")

    # Test rooted union when sets are already connected
    assert not ds.union_left("a", "b")  # Should return False
    assert ds.find("b") == "a"  # Still connected

    # Test error when representative is not a root
    ds.union_left("b", "d")  # "b" is not a root, "a" is its root
    assert ds.find("d") == "a"
    assert ds.connected("b", "d")

    # Test that the specified representative becomes the root
    assert ds.union_left("e", "f")  # "e" becomes root of its own set with "f"
    assert ds.find("e") == "e"
    assert ds.find("f") == "e"

    # Now union the two sets with "a" as the specified root
    assert ds.union_left("a", "e")
    assert ds.find("e") == "a"  # "e"'s set is now under "a"
    assert ds.find("f") == "a"  # "f" also points to "a"
    assert ds.connected("a", "f")

    # Test with KeyError for non-existent values
    with pytest.raises(KeyError):
        ds.union_left("a", "nonexistent")

    with pytest.raises(KeyError):
        ds.union_left("nonexistent", "a")



from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import arith, eqsat, eqsat_pdl_interp, pdl, pdl_interp, test
from xdsl.dialects.builtin import Builtin, FloatAttr, ModuleOp, f32
from xdsl.ir import Block, Region
from xdsl.parser import Parser
from xdsl.transforms.apply_eqsat_pdl_interp import apply_eqsat_pdl_interp


# Surprisingly hard to find a small input and set of patterns that take more than one iteration
# This is due to the union find folding things aggressively as they are rewritten
def test_callback():
    op = ModuleOp(Region(Block()))
    with ImplicitBuilder(op.body):
        c1 = arith.ConstantOp(FloatAttr(1, f32))
        c1_c = eqsat.EClassOp(c1.result)
        add = arith.AddfOp(c1_c, c1_c)
        add_c = eqsat.EClassOp(add.result)
        add2 = arith.AddfOp(add_c, c1_c)
        add2_c = eqsat.EClassOp(add2.result)
        test.TestOp(operands=(add2_c.result,))

    # pattern.mlir
    # pdl.pattern : benefit(1) {
    #     %type = pdl.type : f32

    #     %c1_attr = pdl.attribute = 1.0 : f32
    #     %c1_op = pdl.operation "arith.constant" {"value"=%c1_attr} -> (%type : !pdl.type)
    #     %c1_val = pdl.result 0 of %c1_op

    #     %add_op = pdl.operation "arith.addf" (%c1_val, %c1_val : !pdl.value, !pdl.value) -> (%type : !pdl.type)

    #     pdl.rewrite %add_op {
    #         %c2_attr = pdl.attribute = 2.0 : f32
    #         %c2_op = pdl.operation "arith.constant" {"value"=%c2_attr} -> (%type : !pdl.type)
    #         pdl.replace %add_op with %c2_op
    #     }
    # }

    # // 2 + 1 -> 3
    # pdl.pattern : benefit(1) {
    #     %type = pdl.type : f32

    #     %c2_attr = pdl.attribute = 2.0 : f32
    #     %c2_op = pdl.operation "arith.constant" {"value"=%c2_attr} -> (%type : !pdl.type)
    #     %c2_val = pdl.result 0 of %c2_op

    #     %c1_attr = pdl.attribute = 1.0 : f32
    #     %c1_op = pdl.operation "arith.constant" {"value"=%c1_attr} -> (%type : !pdl.type)
    #     %c1_val = pdl.result 0 of %c1_op

    #     %add_op = pdl.operation "arith.addf" (%c2_val, %c1_val : !pdl.value, !pdl.value) -> (%type : !pdl.type)

    #     pdl.rewrite %add_op {
    #         %c3_attr = pdl.attribute = 3.0 : f32
    #         %c3_op = pdl.operation "arith.constant" {"value"=%c3_attr} -> (%type : !pdl.type)
    #         pdl.replace %add_op with %c3_op
    #     }
    # }

    # // 3 + 1 -> 4
    # pdl.pattern : benefit(1) {
    #     %type = pdl.type : f32

    #     %c3_attr = pdl.attribute = 3.0 : f32
    #     %c3_op = pdl.operation "arith.constant" {"value"=%c3_attr} -> (%type : !pdl.type)
    #     %c3_val = pdl.result 0 of %c3_op

    #     %c1_attr = pdl.attribute = 1.0 : f32
    #     %c1_op = pdl.operation "arith.constant" {"value"=%c1_attr} -> (%type : !pdl.type)
    #     %c1_val = pdl.result 0 of %c1_op

    #     %add_op = pdl.operation "arith.addf" (%c3_val, %c1_val : !pdl.value, !pdl.value) -> (%type : !pdl.type)

    #     pdl.rewrite %add_op {
    #         %c4_attr = pdl.attribute = 4.0 : f32
    #         %c4_op = pdl.operation "arith.constant" {"value"=%c4_attr} -> (%type : !pdl.type)
    #         pdl.replace %add_op with %c4_op
    #     }
    # }

    # xdsl-opt pattern.mlir -p eqsat-pdl-to-pdl-interp
    module_str = """
builtin.module {
  pdl_interp.func @matcher(%0 : !pdl.operation) {
    pdl_interp.check_operation_name of %0 is "arith.addf" -> ^bb0, ^bb1
  ^bb1:
    pdl_interp.finalize
  ^bb0:
    pdl_interp.check_operand_count of %0 is 2 -> ^bb2, ^bb1
  ^bb2:
    pdl_interp.check_result_count of %0 is 1 -> ^bb3, ^bb1
  ^bb3:
    %1 = pdl_interp.get_operand 0 of %0
    pdl_interp.is_not_null %1 : !pdl.value -> ^bb4, ^bb1
  ^bb4:
    %2 = pdl_interp.get_result 0 of %0
    pdl_interp.is_not_null %2 : !pdl.value -> ^bb5, ^bb1
  ^bb5:
    %3 = pdl_interp.get_operand 1 of %0
    pdl_interp.is_not_null %3 : !pdl.value -> ^bb6, ^bb7
  ^bb7:
    eqsat_pdl_interp.choose from ^bb8 then ^bb1
  ^bb8:
    %4 = pdl_interp.get_defining_op of %1 : !pdl.value {position = "root.operand[0].defining_op"}
    pdl_interp.is_not_null %4 : !pdl.operation -> ^bb9, ^bb1
  ^bb9:
    pdl_interp.check_operation_name of %4 is "arith.constant" -> ^bb10, ^bb1
  ^bb10:
    pdl_interp.check_operand_count of %4 is 0 -> ^bb11, ^bb1
  ^bb11:
    pdl_interp.check_result_count of %4 is 1 -> ^bb12, ^bb1
  ^bb12:
    %5 = pdl_interp.get_attribute "value" of %4
    pdl_interp.is_not_null %5 : !pdl.attribute -> ^bb13, ^bb1
  ^bb13:
    pdl_interp.check_attribute %5 is 1.000000e+00 : f32 -> ^bb14, ^bb1
  ^bb14:
    %6 = pdl_interp.get_result 0 of %4
    pdl_interp.is_not_null %6 : !pdl.value -> ^bb15, ^bb1
  ^bb15:
    pdl_interp.are_equal %6, %1 : !pdl.value -> ^bb16, ^bb1
  ^bb16:
    %7 = pdl_interp.get_value_type of %6 : !pdl.type
    %8 = pdl_interp.get_value_type of %2 : !pdl.type
    pdl_interp.are_equal %7, %8 : !pdl.type -> ^bb17, ^bb1
  ^bb17:
    pdl_interp.check_type %7 is f32 -> ^bb18, ^bb1
  ^bb18:
    pdl_interp.record_match @rewriters::@pdl_generated_rewriter(%0 : !pdl.operation) : benefit(1), loc([]), root("arith.addf") -> ^bb1
  ^bb6:
    eqsat_pdl_interp.choose from ^bb19 then ^bb7
  ^bb19:
    %9 = pdl_interp.get_defining_op of %1 : !pdl.value {position = "root.operand[0].defining_op"}
    pdl_interp.is_not_null %9 : !pdl.operation -> ^bb20, ^bb1
  ^bb20:
    pdl_interp.check_operation_name of %9 is "arith.constant" -> ^bb21, ^bb1
  ^bb21:
    pdl_interp.check_operand_count of %9 is 0 -> ^bb22, ^bb1
  ^bb22:
    pdl_interp.check_result_count of %9 is 1 -> ^bb23, ^bb1
  ^bb23:
    %10 = pdl_interp.get_attribute "value" of %9
    pdl_interp.is_not_null %10 : !pdl.attribute -> ^bb24, ^bb1
  ^bb24:
    pdl_interp.switch_attribute %10 to [2.000000e+00 : f32, 3.000000e+00 : f32](^bb25, ^bb26) -> ^bb1
  ^bb25:
    %11 = pdl_interp.get_result 0 of %9
    pdl_interp.is_not_null %11 : !pdl.value -> ^bb27, ^bb1
  ^bb27:
    pdl_interp.are_equal %11, %1 : !pdl.value -> ^bb28, ^bb1
  ^bb28:
    %12 = pdl_interp.get_value_type of %11 : !pdl.type
    %13 = pdl_interp.get_value_type of %2 : !pdl.type
    pdl_interp.are_equal %12, %13 : !pdl.type -> ^bb29, ^bb1
  ^bb29:
    pdl_interp.check_type %12 is f32 -> ^bb30, ^bb1
  ^bb30:
    eqsat_pdl_interp.choose from ^bb31 then ^bb1
  ^bb31:
    %14 = pdl_interp.get_defining_op of %3 : !pdl.value {position = "root.operand[1].defining_op"}
    pdl_interp.is_not_null %14 : !pdl.operation -> ^bb32, ^bb1
  ^bb32:
    pdl_interp.check_operation_name of %14 is "arith.constant" -> ^bb33, ^bb1
  ^bb33:
    pdl_interp.check_operand_count of %14 is 0 -> ^bb34, ^bb1
  ^bb34:
    pdl_interp.check_result_count of %14 is 1 -> ^bb35, ^bb1
  ^bb35:
    %15 = pdl_interp.get_attribute "value" of %14
    pdl_interp.is_not_null %15 : !pdl.attribute -> ^bb36, ^bb1
  ^bb36:
    pdl_interp.check_attribute %15 is 1.000000e+00 : f32 -> ^bb37, ^bb1
  ^bb37:
    %16 = pdl_interp.get_result 0 of %14
    pdl_interp.is_not_null %16 : !pdl.value -> ^bb38, ^bb1
  ^bb38:
    pdl_interp.are_equal %16, %3 : !pdl.value -> ^bb39, ^bb1
  ^bb39:
    %17 = pdl_interp.get_value_type of %16 : !pdl.type
    pdl_interp.are_equal %12, %17 : !pdl.type -> ^bb40, ^bb1
  ^bb40:
    pdl_interp.record_match @rewriters::@pdl_generated_rewriter_2(%0 : !pdl.operation) : benefit(1), loc([]), root("arith.addf") -> ^bb1
  ^bb26:
    %18 = pdl_interp.get_result 0 of %9
    pdl_interp.is_not_null %18 : !pdl.value -> ^bb41, ^bb1
  ^bb41:
    pdl_interp.are_equal %18, %1 : !pdl.value -> ^bb42, ^bb1
  ^bb42:
    %19 = pdl_interp.get_value_type of %18 : !pdl.type
    %20 = pdl_interp.get_value_type of %2 : !pdl.type
    pdl_interp.are_equal %19, %20 : !pdl.type -> ^bb43, ^bb1
  ^bb43:
    pdl_interp.check_type %19 is f32 -> ^bb44, ^bb1
  ^bb44:
    eqsat_pdl_interp.choose from ^bb45 then ^bb1
  ^bb45:
    %21 = pdl_interp.get_defining_op of %3 : !pdl.value {position = "root.operand[1].defining_op"}
    pdl_interp.is_not_null %21 : !pdl.operation -> ^bb46, ^bb1
  ^bb46:
    pdl_interp.check_operation_name of %21 is "arith.constant" -> ^bb47, ^bb1
  ^bb47:
    pdl_interp.check_operand_count of %21 is 0 -> ^bb48, ^bb1
  ^bb48:
    pdl_interp.check_result_count of %21 is 1 -> ^bb49, ^bb1
  ^bb49:
    %22 = pdl_interp.get_attribute "value" of %21
    pdl_interp.is_not_null %22 : !pdl.attribute -> ^bb50, ^bb1
  ^bb50:
    pdl_interp.check_attribute %22 is 1.000000e+00 : f32 -> ^bb51, ^bb1
  ^bb51:
    %23 = pdl_interp.get_result 0 of %21
    pdl_interp.is_not_null %23 : !pdl.value -> ^bb52, ^bb1
  ^bb52:
    pdl_interp.are_equal %23, %3 : !pdl.value -> ^bb53, ^bb1
  ^bb53:
    %24 = pdl_interp.get_value_type of %23 : !pdl.type
    pdl_interp.are_equal %19, %24 : !pdl.type -> ^bb54, ^bb1
  ^bb54:
    pdl_interp.record_match @rewriters::@pdl_generated_rewriter_3(%0 : !pdl.operation) : benefit(1), loc([]), root("arith.addf") -> ^bb1
  }
  builtin.module @rewriters {
    pdl_interp.func @pdl_generated_rewriter(%0 : !pdl.operation) {
      %1 = pdl_interp.create_attribute 2.000000e+00 : f32
      %2 = pdl_interp.create_type f32
      %3 = pdl_interp.create_operation "arith.constant" {"value" = %1} -> (%2 : !pdl.type)
      %4 = pdl_interp.get_results of %3 : !pdl.range<value>
      pdl_interp.replace %0 with (%4 : !pdl.range<value>)
      pdl_interp.finalize
    }
    pdl_interp.func @pdl_generated_rewriter_2(%0 : !pdl.operation) {
      %1 = pdl_interp.create_attribute 3.000000e+00 : f32
      %2 = pdl_interp.create_type f32
      %3 = pdl_interp.create_operation "arith.constant" {"value" = %1} -> (%2 : !pdl.type)
      %4 = pdl_interp.get_results of %3 : !pdl.range<value>
      pdl_interp.replace %0 with (%4 : !pdl.range<value>)
      pdl_interp.finalize
    }
    pdl_interp.func @pdl_generated_rewriter_3(%0 : !pdl.operation) {
      %1 = pdl_interp.create_attribute 4.000000e+00 : f32
      %2 = pdl_interp.create_type f32
      %3 = pdl_interp.create_operation "arith.constant" {"value" = %1} -> (%2 : !pdl.type)
      %4 = pdl_interp.get_results of %3 : !pdl.range<value>
      pdl_interp.replace %0 with (%4 : !pdl.range<value>)
      pdl_interp.finalize
    }
  }
}
    """
    ctx = Context()
    ctx.load_dialect(pdl_interp.PDLInterp)
    ctx.load_dialect(arith.Arith)
    ctx.load_dialect(eqsat.EqSat)
    ctx.load_dialect(pdl.PDL)
    ctx.load_dialect(eqsat_pdl_interp.EqSatPDLInterp)
    ctx.load_dialect(Builtin)
    ctx.load_dialect(test.Test)
    patterns_module = Parser(ctx, module_str).parse_module()

    strs: list[str] = []

    def callback(module: ModuleOp):
        strs.append(str(module))

    apply_eqsat_pdl_interp(op, ctx, patterns_module, 10, callback)

    assert strs == [
        """\
builtin.module {
  %0 = arith.constant 1.000000e+00 : f32
  %1 = eqsat.eclass %0 : f32
  %2 = arith.constant 2.000000e+00 : f32
  %3 = eqsat.const_eclass %2, %4 (constant = 2.000000e+00 : f32) : f32
  %4 = arith.addf %1, %1 : f32
  %5 = arith.addf %3, %1 : f32
  %6 = eqsat.eclass %5 : f32
  "test.op"(%6) : (f32) -> ()
}""",
        """\
builtin.module {
  %0 = arith.constant 1.000000e+00 : f32
  %1 = eqsat.eclass %0 : f32
  %2 = arith.constant 2.000000e+00 : f32
  %3 = eqsat.const_eclass %2, %4 (constant = 2.000000e+00 : f32) : f32
  %4 = arith.addf %1, %1 : f32
  %5 = arith.constant 3.000000e+00 : f32
  %6 = eqsat.const_eclass %5, %7 (constant = 3.000000e+00 : f32) : f32
  %7 = arith.addf %3, %1 : f32
  "test.op"(%6) : (f32) -> ()
}""",
    ]
